/* -*-C++-*- */
/*
 * Domain descriptions.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: domains.h,v 2.2 2005/02/08 09:25:38 lorens Exp $
 */
#ifndef DOMAINS_H
#define DOMAINS_H

#include <config.h>
#include "events.h"
#include "functions.h"
#include "predicates.h"
#include "terms.h"
#include "types.h"
#include <iostream>
#include <map>
#include <string>

struct Problem;


/* ====================================================================== */
/* Domain */

/*
 * Domain definition.
 */
struct Domain {
  /* Table of domain definitions. */
  struct DomainMap : public std::map<std::string, const Domain*> {
  };

  /* Returns a const_iterator pointing to the first domain. */
  static DomainMap::const_iterator begin();

  /* Returns a const_iterator pointing beyond the last domain. */
  static DomainMap::const_iterator end();

  /* Returns a pointer to the domain with the given name, or 0 if it
     is undefined. */
  static const Domain* find(const std::string& name);

  /* Removes all defined domains. */
  static void clear();

  /* Constructs an empty domain with the given name. */
  Domain(const std::string& name);

  /* Deletes a domain. */
  ~Domain();

  /* Returns the name of this domain. */
  const std::string& name() const { return name_; }

  /* Returns the type table of this domain. */
  TypeTable& types() { return types_; }

  /* Returns the type table of this domain. */
  const TypeTable& types() const { return types_; }

  /* Returns the predicate table of this domain. */
  PredicateTable& predicates() { return predicates_; }

  /* Returns the predicate table of this domain. */
  const PredicateTable& predicates() const { return predicates_; }

  /* Returns the function table of this domain. */
  FunctionTable& functions() { return functions_; }

  /* Returns the function table of this domain. */
  const FunctionTable& functions() const { return functions_; }

  /* Returns the `total-time' function. */
  const Function& total_time() const { return total_time_; }

  /* Returns the `goal-achieved' function. */
  const Function& goal_achieved() const { return goal_achieved_; }

  /* Returns the term table of this domain. */
  TermTable& terms() { return terms_; }

  /* Returns the term table of this domain. */
  const TermTable& terms() const { return terms_; }

  /* Adds the given event to this domain. */
  void add_event(const EventSchema& event);

  /* Returns a pointer to the event with the given name, or 0 if there
     is no event with the given name. */
  const EventSchema* find_event(const std::string& name) const;

  /* Returns the events of this domain. */
  const EventSchemaMap& events() const { return events_; }

private:
  /* Table of all defined domains. */
  static DomainMap domains;

  /* Name of this domain. */
  std::string name_;
  /* Domain types. */
  TypeTable types_;
  /* Domain predicates. */
  PredicateTable predicates_;
  /* Domain functions. */
  FunctionTable functions_;
  /* The `total-time' function. */
  Function total_time_;
  /* The `goal-achieved' function. */
  Function goal_achieved_;
  /* Domain terms. */
  TermTable terms_;
  /* Domain events. */
  EventSchemaMap events_;
};

/* Output operator for domains. */
std::ostream& operator<<(std::ostream& os, const Domain& d);


#endif /* DOMAINS_H */
