/*
 * Events.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: events.cc,v 2.2 2005/02/08 11:16:51 lorens Exp $
 */
#include "events.h"
#include <stack>


/* ====================================================================== */
/* EventSchema */

/* Constructs an event schema with the given name. */
EventSchema::EventSchema(const std::string& name, bool controllable)
  : name_(name), controllable_(controllable),
    condition_(&StateFormula::TRUE), delay_(&Distribution::EXP1),
    effect_(&Effect::EMPTY) {
  RCObject::ref(condition_);
  RCObject::ref(delay_);
  RCObject::ref(effect_);
}


/* Deletes this event schema. */
EventSchema::~EventSchema() {
  RCObject::destructive_deref(condition_);
  RCObject::destructive_deref(delay_);
  RCObject::destructive_deref(effect_);
}


/* Sets the condition of this event schema. */
void EventSchema::set_condition(const StateFormula& condition) {
  if (&condition != condition_) {
    RCObject::ref(&condition);
    RCObject::destructive_deref(condition_);
    condition_ = &condition;
  }
}


/* Sets the delay distribution of this event schema. */
void EventSchema::set_delay(const Distribution& delay) {
  if (&delay != delay_) {
    RCObject::ref(&delay);
    RCObject::destructive_deref(delay_);
    delay_ = &delay;
  }
}


/* Sets the effect of this event schema. */
void EventSchema::set_effect(const Effect& effect) {
  if (&effect != effect_) {
    RCObject::ref(&effect);
    RCObject::destructive_deref(effect_);
    effect_ = &effect;
  }
}


/* Fills the provided list with instantiations of this event
   schema. */
void EventSchema::instantiations(EventSet& events, const TermTable& terms,
				 const AtomSet& atoms,
				 const ValueMap& values) const {
  size_t n = parameters().size();
  if (n == 0) {
    const StateFormula& cond =
      condition().instantiation(SubstitutionMap(), terms, atoms, values);
    if (!cond.contradiction()) {
      events.insert(&instantiation(SubstitutionMap(), terms,
				   atoms, values, cond));
    }
  } else {
    SubstitutionMap args;
    std::vector<const ObjectList*> arguments(n);
    std::vector<ObjectList::const_iterator> next_arg;
    for (size_t i = 0; i < n; i++) {
      Type t = TermTable::type(parameters()[i]);
      arguments[i] = &terms.compatible_objects(t);
      if (arguments[i]->empty()) {
	return;
      }
      next_arg.push_back(arguments[i]->begin());
    }
    std::stack<const StateFormula*> conds;
    conds.push(&condition());
    RCObject::ref(conds.top());
    for (size_t i = 0; i < n; ) {
      args.insert(std::make_pair(parameters()[i], *next_arg[i]));
      SubstitutionMap pargs;
      pargs.insert(std::make_pair(parameters()[i], *next_arg[i]));
      const StateFormula& cond =
	conds.top()->instantiation(pargs, terms, atoms, values);
      conds.push(&cond);
      RCObject::ref(conds.top());
      if (i + 1 == n || cond.contradiction()) {
	if (!cond.contradiction()) {
	  events.insert(&instantiation(args, terms, atoms, values, cond));
	}
	for (int j = i; j >= 0; j--) {
	  RCObject::destructive_deref(conds.top());
	  conds.pop();
	  args.erase(parameters()[j]);
	  next_arg[j]++;
	  if (next_arg[j] == arguments[j]->end()) {
	    if (j == 0) {
	      i = n;
	      break;
	    } else {
	      next_arg[j] = arguments[j]->begin();
	    }
	  } else {
	    i = j;
	    break;
	  }
	}
      } else {
	i++;
      }
    }
    while (!conds.empty()) {
      RCObject::destructive_deref(conds.top());
      conds.pop();
    }
  }
}


/* Returns an instantiation of this event schema. */
const Event& EventSchema::instantiation(const SubstitutionMap& subst,
					const TermTable& terms,
					const AtomSet& atoms,
					const ValueMap& values,
					const StateFormula& cond) const {
  Event* event = new Event(name(), controllable());
  for (VariableList::const_iterator vi = parameters().begin();
       vi != parameters().end(); vi++) {
    SubstitutionMap::const_iterator si = subst.find(*vi);
    event->add_argument((*si).second);
  }
  event->set_condition(cond);
  event->set_delay(delay());
  event->set_effect(effect().instantiation(subst, terms, atoms, values));
  return *event;
}


/* Output operator for event schemas. */
std::ostream& operator<<(std::ostream& os, const EventSchema& e) {
  os << "  " << e.name();
  if (e.controllable()) {
    os << " <action>";
  }
  os << std::endl << "    parameters:";
  for (VariableList::const_iterator vi = e.parameters().begin();
       vi != e.parameters().end(); vi++) {
    os << ' ' << *vi;
  }
  os << std::endl << "    condition: " << e.condition();
  os << std::endl << "    delay: " << e.delay();
  os << std::endl << "    effect:" << e.effect();
  return os;
}


/* ====================================================================== */
/* Event */

/* Constructs an event with the given name. */
Event::Event(const std::string& name, bool controllable)
  : name_(name), controllable_(controllable),
    condition_(&StateFormula::TRUE), delay_(&Distribution::EXP1),
    effect_(&Effect::EMPTY) {
  RCObject::ref(condition_);
  RCObject::ref(delay_);
  RCObject::ref(effect_);
}


/* Deletes this event. */
Event::~Event() {
  RCObject::destructive_deref(condition_);
  RCObject::destructive_deref(delay_);
  RCObject::destructive_deref(effect_);
}


/* Sets the condition of this event. */
void Event::set_condition(const StateFormula& condition) {
  if (&condition != condition_) {
    RCObject::ref(&condition);
    RCObject::destructive_deref(condition_);
    condition_ = &condition;
  }
}


/* Sets the delay distribution of this event. */
void Event::set_delay(const Distribution& delay) {
  if (&delay != delay_) {
    RCObject::ref(&delay);
    RCObject::destructive_deref(delay_);
    delay_ = &delay;
  }
}


/* Sets the effect of this event. */
void Event::set_effect(const Effect& effect) {
  if (&effect != effect_) {
    RCObject::ref(&effect);
    RCObject::destructive_deref(effect_);
    effect_ = &effect;
  }
}


/* Tests if this event is enabled in the given state. */
bool Event::enabled(const TermTable& terms,
		    const AtomSet& atoms, const ValueMap& values) const {
  return condition().holds(atoms, values);
}


/* Changes the given state according to the effects of this event. */
void Event::affect(AtomSet& atoms, ValueMap& values) const {
  AtomList adds;
  AtomList deletes;
  UpdateList updates;
  effect().state_change(adds, deletes, updates, atoms, values);
  for (AtomList::const_iterator ai = deletes.begin();
       ai != deletes.end(); ai++) {
    atoms.erase(*ai);
  }
  atoms.insert(adds.begin(), adds.end());
  for (UpdateList::const_iterator ui = updates.begin();
       ui != updates.end(); ui++) {
    (*ui)->affect(values);
  }
}


/* Output operator for events. */
std::ostream& operator<<(std::ostream& os, const Event& e) {
  os << '(' << e.name();
  for (ObjectList::const_iterator oi = e.arguments().begin();
       oi != e.arguments().end(); oi++) {
    os << ' ' << *oi;
  }
  os << ')';
  return os;
}
