/* -*-C++-*- */
/*
 * Events.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: events.h,v 2.2 2005/02/08 11:18:10 lorens Exp $
 */
#ifndef EVENTS_H
#define EVENTS_H

#include <config.h>
#include "effects.h"
#include "formulas.h"
#include "expressions.h"
#include "distributions.h"
#include "terms.h"
#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>


/* ====================================================================== */
/* EventSchema */

struct Event;
struct EventSet;

/*
 * Event schema.
 */
struct EventSchema {
  /* Constructs an event schema with the given name. */
  EventSchema(const std::string& name, bool controllable);

  /* Deletes this event schema. */
  ~EventSchema();

  /* Adds a parameter to this event schema. */
  void add_parameter(const Variable& parameter) {
    parameters_.push_back(parameter);
  }

  /* Sets the condition of this event schema. */
  void set_condition(const StateFormula& condition);

  /* Sets the delay distribution of this event schema. */
  void set_delay(const Distribution& delay);

  /* Sets the effect of this event schema. */
  void set_effect(const Effect& effect);
  
  /* Returns the name of this event schema. */
  const std::string& name() const { return name_; }

  /* Tests if this event schema is controllable. */
  bool controllable() const { return controllable_; }

  /* Returns the parameters of this event schema. */
  const VariableList& parameters() const { return parameters_; }

  /* Returns the conditions of this event schema. */
  const StateFormula& condition() const { return *condition_; }

  /* Returns the delay distribution of this event schema. */
  const Distribution& delay() const { return *delay_; }

  /* Returns the effect of this event schema. */
  const Effect& effect() const { return *effect_; }

  /* Fills the provided list with instantiations of this event
     schema. */
  void instantiations(EventSet& events, const TermTable& terms,
		      const AtomSet& atoms, const ValueMap& values) const;

private:
  /* Event name. */
  std::string name_;
  /* Whether this is a controllable event or not. */
  bool controllable_;
  /* Event parameters. */
  VariableList parameters_;
  /* Event condition. */
  const StateFormula* condition_;
  /* Event delay distribution. */
  const Distribution* delay_;
  /* Event effects. */
  const Effect* effect_;

  /* Returns an instantiation of this event schema. */
  const Event& instantiation(const SubstitutionMap& subst,
			     const TermTable& terms,
			     const AtomSet& atoms,
			     const ValueMap& values,
			     const StateFormula& precond) const;
};

/* Output operator for event schemas. */
std::ostream& operator<<(std::ostream& os, const EventSchema& e);


/* ====================================================================== */
/* EventSchemaMap */

/*
 * Table of event schemas.
 */
struct EventSchemaMap : public std::map<std::string, const EventSchema*> {
};


/* ====================================================================== */
/* Event */

/*
 * A fully instantiated event.
 */
struct Event {
  /* Constructs an event with the given name. */
  Event(const std::string& name, bool controllable);

  /* Deletes this event. */
  ~Event();

  /* Adds an argument to this event. */
  void add_argument(const Object& argument) { arguments_.push_back(argument); }

  /* Sets the condition of this event. */
  void set_condition(const StateFormula& condition);

  /* Sets the delay distribution of this event. */
  void set_delay(const Distribution& delay);

  /* Sets the effect of this event. */
  void set_effect(const Effect& effect);
  
  /* Returns the name of this event. */
  const std::string& name() const { return name_; }

  /* Tests if this event is controllable. */
  bool controllable() const { return controllable_; }

  /* Returns the arguments of this event. */
  const ObjectList& arguments() const { return arguments_; }

  /* Returns the condition of this event. */
  const StateFormula& condition() const { return *condition_; }

  /* Returns the delay distribution of this event. */
  const Distribution& delay() const { return *delay_; }

  /* Returns the effect of this event. */
  const Effect& effect() const { return *effect_; }

  /* Tests if this event is enabled in the given state. */
  bool enabled(const TermTable& terms,
	       const AtomSet& atoms, const ValueMap& values) const;

  /* Changes the given state according to the effects of this event. */
  void affect(AtomSet& atoms, ValueMap& values) const;

private:
  /* Event name. */
  std::string name_;
  /* Whether this is a controllable event or not. */
  bool controllable_;
  /* Event arguments. */
  ObjectList arguments_;
  /* Event condition. */
  const StateFormula* condition_;
  /* Event delay distribution. */
  const Distribution* delay_;
  /* Event effects. */
  const Effect* effect_;
};

/* Output operator for events. */
std::ostream& operator<<(std::ostream& os, const Event& e);

/*
 * Less than function object for event pointers.
 */
namespace std {
  struct less<const Event*>
    : public binary_function<const Event*, const Event*, bool> {
    /* Comparison function operator. */
    bool operator()(const Event* e1, const Event* e2) const {
      if (e1 == 0) {
	return e2 != 0;
      } else if (e2 == 0) {
	return false;
      } else if (e1->name() < e2->name()) {
	return true;
      } else if (e2->name() < e1->name()) {
	return false;
      } else {
	return e1->arguments() < e2->arguments();
      }
    }
  };
}


/* ====================================================================== */
/* EventSet */

/*
 * Set of events.
 */
struct EventSet : public std::set<const Event*> {
};


/* ====================================================================== */
/* EventList */

/*
 * List of events.
 */
struct EventList : public std::vector<const Event*> {
};


#endif /* EVENTS_H */
