/*
 * Moments matching.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: moments.cc,v 2.1 2005/02/08 11:26:06 lorens Exp $
 */
#include "phfit.h"
#include <stdexcept>


/* Provides the parameters for a continuous phase-type distribution
   matching the first moment of a distribution. */
static void ph1(PHParameters& params, double mu1) {
  params.n = 1;
  params.add_rate(0, 1, 1.0/mu1);
}


/* Provides the parameters for a continuous phase-type distribution
   matching the first two moments of a distribution. */
static void ph2(PHParameters& params, double mu1, double mu2) {
  double cv2 = mu2/(mu1*mu1) - 1.0;
  if (cv2 >= 0.5) {
    params.n = 2;
    double p = 1/(2.0*cv2);
    double lambda1 = 2.0/mu1;
    double lambda2 = 1.0/(mu1*cv2);
    params.add_rate(0, 1, p*lambda1);
    if (fabs(1.0 - p) >= 1e-15) {
      params.add_rate(0, 2, (1.0 - p)*lambda1);
    }
    params.add_rate(1, 2, lambda2);
  } else {
    double n = 1.0/cv2;
    if (fabs(round(n) - n) < 1e-15) {
      n = round(n);
    } else {
      n = ceil(n);
    }
    params.n = int(n + 0.5);
    double p = 1.0 - ((2.0*n*cv2 + n - 2.0 - sqrt(n*n + 4.0 - 4.0*n*cv2))
		      /2.0/(n - 1.0)/(cv2 + 1.0));
    double lambda = (1.0 - p + n*p)/mu1;
    params.add_rate(0, 1, p*lambda);
    if (fabs(1.0 - p) >= 1e-15) {
      params.add_rate(0, params.n, (1.0 - p)*lambda);
    }
    for (size_t i = 1; i < params.n; i++) {
      params.add_rate(i, i + 1, lambda);
    }
  }
}


/* Provides the parameters for a continuous phase-type distribution
   matching the first three moments of a distribution. */
static void ph3(PHParameters& params, double mu1, double mu2, double mu3) {
  /*
   * Implements the moment matching algorithm described in:
   *
   *   Osogami, Takayuki, and Mor Harchol-Balter.  2003.  "A
   *     closed-form solution for mapping general distributions to
   *     minimal PH distributions".  In Proceedings of the 13th
   *     International Conference on Tools and Algorithms for the
   *     Construnction and Analysis of Systems.
   *
   * This is a slightly modified version avoiding solutions with p<1,
   * at the price that some positive distributions cannot be properly
   * represented.  For a distribution that the original algorithm
   * would map to an EC distribution with p<1, this algorithm modifies
   * the second and third moments so that the modified distribution
   * can be matched to an EC distribution with p=1.  This is to avoid
   * EC distributions with mass probability at zero, which do not fit
   * well with the model of discrete event systems that we are using.
   */
  double m2G = mu2/(mu1*mu1);
  double m3G = mu3/(mu1*mu2);
  double x = 1.0/(m2G - 1.0);
  if (fabs(round(x) - x) < 1e-15) {
    /*
     * Adjust second moment so that 1/(m2G - 1) is not an integer.
     */
    m2G *= (1.0 - 1e-5);
    std::cerr << std::endl
	      << PACKAGE ": second moment changed from " << mu2
	      << " to " << m2G*mu1*mu1 << std::endl;
  }
  if (m3G < 2.0*m2G - 1.0) {
    /*
     * Adjust third moment so that the resulting EC distribution has
     * no mass probability at zero (p=1).
     */
    m3G = 2.0*m2G - 1.0;
    std::cerr << std::endl
	      << PACKAGE ": third moment changed from " << mu3
	      << " to " << m3G*mu1*mu2 << std::endl;
  }
  double n;
  if (m2G < 2.0 && fabs(m3G - (2.0*m2G - 1.0)) < 1e-15) {
    n = floor(m2G/(m2G - 1.0));
  } else {
    n = floor(m2G/(m2G - 1.0) + 1.0);
  }
  double m2X = ((n - 3.0)*m2G - (n - 2.0))/((n - 2.0)*m2G - (n - 1.0));
  double mu1X = mu1/((n - 2.0)*m2X - (n - 3.0));
  double alpha = (n - 2.0)*(m2X - 1.0)*(n*(n - 1.0)*m2X*m2X
					- n*(2.0*n - 5.0)*m2X
					+ (n - 1.0)*(n - 3.0));
  double tmp = (n - 2.0)*m2X - (n - 3.0);
  double beta = ((n - 1.0)*m2X - (n - 2.0))*tmp*tmp;
  double m3X = (beta*m3G - alpha)/m2X;
  double u, v;
  if (fabs(3.0*m2X - 2.0*m3X) < 1e-15) {
    u = 1.0;
    v = 0.0;
  } else {
    u = (6.0 - 2.0*m3X)/(3.0*m2X - 2.0*m3X);
    v = (12.0 - 6.0*m2X)/(m2X*(3.0*m2X - 2.0*m3X));
  }
  params.n = int(n + 0.5);
  double lambda_x = 1.0/((m2X - 1.0)*mu1X);
  for (size_t i = 0; i < params.n - 1; i++) {
    params.add_rate(i, i + 1, lambda_x);
  }
  tmp = sqrt(u*u - 4.0*v);
  double lambda_x1 = (u + tmp)/(2.0*mu1X);
  if (fabs(lambda_x1*mu1X - 1.0) < 1e-15) {
    params.n--;
    params.add_rate(params.n - 1, params.n, lambda_x1);
  } else {
    double lambda_x2 = (u - tmp)/(2.0*mu1X);
    double p_x = lambda_x2*(lambda_x1*mu1X - 1.0)/lambda_x1;
    params.add_rate(params.n - 2, params.n - 1, p_x*lambda_x1);
    params.add_rate(params.n - 2, params.n, (1.0 - p_x)*lambda_x1);
    params.add_rate(params.n - 1, params.n, lambda_x2);
  }
}


/* Provides the parameters for a continuous phase-type distribution
   matching the first n moments of this distribution. */
void ph_matching_moments(PHParameters& params,
			 const Distribution& dist, size_t n) {
  std::vector<double> m;
  dist.moments(m, n);
  switch (n) {
  case 1:
    ph1(params, m[0]);
    break;
  case 2:
    ph2(params, m[0], m[1]);
    break;
  case 3:
    ph3(params, m[0], m[1], m[2]);
    break;
  default:
    throw std::out_of_range("moments matching for phase-type distributions is"
			    " limited to 1, 2, or 3 moments");
  }
}
