/*
 * MTBDD representation.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: mtbdd.cc,v 2.3 2005/02/08 11:35:22 lorens Exp $
 */
#include "mtbdd.h"
#include "states.h"
#include "formulas.h"
#include "functions.h"
#include "phfit.h"
#include <map>
#include <stdexcept>
#include <sys/time.h>


/*
 * Data for phase-type distribution.
 */
struct PHData {
  PHParameters params;
  int low_bit;
  int high_bit;
};


/* Verbosity level. */
extern int verbosity;

/* The reward function. */
static const Function* reward_function;
/* State variables for the current problem. */
static std::map<const Atom*, int> state_variables;
/* A mapping from variable indices to atoms. */
static std::map<int, const Atom*> dynamic_atoms;
/* Phase-type distributions for event schemas. */
static PHEventMap ph_parameters;
/* Phase-type distributions for events. */
static std::map<const Event*, PHData> ph_events;
/* Total number of DD variables. */
static int nvars;
/* Number of auxiliary variables. */
static int aux_vars;
/* Number of phase variables. */
static int ph_vars;
/* Variable ordering. */
static std::vector<int> var_order;
/* Ordered variables. */
static std::vector<int> ordered_vars;
/* State holding times. */
std::map<const Event*, DdNode*> holding_times;
/* Event transition matrices. */
static std::multimap<const Event*, DdNode*> event_transitions;
/* Action transition matrices. */
static std::multimap<const Event*, DdNode*> action_transitions;
/* Update filters for action phases. */
static std::map<const Event*, DdNode*> action_phase_updates;
/* Action reward vectors. */
static std::map<const Event*, DdNode*> action_rewards;

typedef std::multimap<const Event*, DdNode*>::const_iterator trans_iter;


/* ====================================================================== */
/* ceil_log2 */

/*
 * Returns the ceiling of the base-2 logarithm of the given integer.
 */
static int ceil_log2(int n) {
  int result = 0;
  for (int i = n - 1; i > 0; i >>= 1) {
    result++;
  }
  return result;
}


/* ====================================================================== */
/* collect_state_variables */

/*
 * Collects state variables from the given effect.
 */
static void collect_state_variables(const Effect& effect,
				    const Domain& domain) {
  if (effect.empty()) {
    /*
     * This effect is empty.
     */
    return;
  }

  const UpdateEffect* ue = dynamic_cast<const UpdateEffect*>(&effect);
  if (ue != 0) {
    /*
     * Only reward updates are supported, and they do not involve any
     * state variables.
     */
    const Function& function = ue->update().fluent().function();
    if (function != domain.total_time() && function != domain.goal_achieved()
	&& (reward_function == 0 || function != *reward_function)) {
      throw std::logic_error("numeric state variables not supported");
    }
    return;
  }

  const SimpleEffect* se = dynamic_cast<const SimpleEffect*>(&effect);
  if (se != 0) {
    /*
     * A simple effect involves a single state variable.
     */
    const Atom* atom = &se->atom();
    if (state_variables.find(atom) == state_variables.end()) {
      dynamic_atoms.insert(std::make_pair(nvars, atom));
      state_variables.insert(std::make_pair(atom, nvars));
      nvars++;
    }
    return;
  }

  const ConjunctiveEffect* ce =
    dynamic_cast<const ConjunctiveEffect*>(&effect);
  if (ce != 0) {
    /*
     * The state variables of a conjunctive effect are the state
     * variables of the conjuncts.
     */
    for (EffectList::const_iterator ei = ce->conjuncts().begin();
	 ei != ce->conjuncts().end(); ei++) {
      collect_state_variables(**ei, domain);
    }
    return;
  }

  const ConditionalEffect* we =
    dynamic_cast<const ConditionalEffect*>(&effect);
  if (we != 0) {
    /*
     * The state variables of a conditional effect are the state
     * variables of the condition and the effect.
     */
    collect_state_variables(we->effect(), domain);
    return;
  }

  const ProbabilisticEffect* pe =
    dynamic_cast<const ProbabilisticEffect*>(&effect);
  if (pe != 0) {
    /*
     * The state variables of a probabilistic effect are the state
     * variables of the possible effects.
     */
    for (size_t i = 0; i < pe->size(); i++) {
      collect_state_variables(pe->effect(i), domain);
    }
    return;
  }

  /*
   * No other types of effects exist.
   */
  throw std::logic_error("unexpected effect");
}


/* ====================================================================== */
/* auxiliary_dbn_variables */

/*
 * Calculates the number of auxiliary DBN variables that are required
 * for the given effect.
 */
static int auxiliary_dbn_variables(const Effect& effect) {
  const ConjunctiveEffect* ce =
    dynamic_cast<const ConjunctiveEffect*>(&effect);
  if (ce != 0) {
    int n = 0;
    for (EffectList::const_iterator ei = ce->conjuncts().begin();
	 ei != ce->conjuncts().end(); ei++) {
      n += auxiliary_dbn_variables(**ei);
    }
    return n;
  }

  const ConditionalEffect* we =
    dynamic_cast<const ConditionalEffect*>(&effect);
  if (we != 0) {
    return auxiliary_dbn_variables(we->effect());
  }

  const ProbabilisticEffect* pe =
    dynamic_cast<const ProbabilisticEffect*>(&effect);
  if (pe != 0) {
    Rational p_rest = 1;
    size_t n = pe->size();
    for (size_t i = 0; i < n; i++) {
      p_rest = p_rest - pe->probability(i);
    }
    if (p_rest > 0) {
      n++;
    }
    n = ceil_log2(n);
    for (size_t i = 0; i < pe->size(); i++) {
      n += auxiliary_dbn_variables(pe->effect(i));
    }
    return n;
  }

  return 0;
}


/*
 * Calculates the number of auxiliary DBN variables that are required
 * for probabilistic effects of the given events.
 */
static int auxiliary_dbn_variables(const EventSet& events) {
  int num_bits = 0;
  for (EventSet::const_iterator ei = events.begin();
       ei != events.end(); ei++) {
    num_bits = std::max(num_bits, auxiliary_dbn_variables((*ei)->effect()));
  }
  return num_bits;
}


/* ====================================================================== */
/* generate_ph_distributions */

/*
 * Generates phase-type distributions for events, and returns the
 * total number of boolean state variables required to represent the
 * phases.
 */
static int generate_ph_distributions(const EventSet& events,
				     size_t moments, size_t phases) {
  int num_bits = 0;
  for (EventSet::const_iterator ei = events.begin();
       ei != events.end(); ei++) {
    const Distribution& dist = (*ei)->delay();
    if (typeid(dist) != typeid(Exponential)) {
      PHData& data = ph_events[*ei];
      PHEventMap::const_iterator di = ph_parameters.find((*ei)->name());
      if (di != ph_parameters.end()) {
	data.params = (*di).second;
      } else {
	if (moments > 0) {
	  ph_matching_moments(data.params, dist, moments);
	} else {
	  ph_em_fit(data.params, dist, phases, 1e-2, 1e-6, 10000);
	}
	if (verbosity > 0) {
	  std::cout << std::endl
		    << "PH distribution for " << dist << ':' << std::endl
		    << "  n = " << data.params.n << std::endl;
	  if (verbosity > 2) {
	    for (TransitionRateMap::const_iterator ri =
		   data.params.rates.begin();
		 ri != data.params.rates.end(); ri++) {
	      std::cout << "  <" << (*ri).first.first << ','
			<< (*ri).first.second << "> = " << (*ri).second
			<< std::endl;
	    }
	  }
	}
	ph_parameters.insert(std::make_pair((*ei)->name(), data.params));
      }
      int n = data.params.n;
      if (n > 1) {
	data.low_bit = num_bits + nvars;
	data.high_bit = data.low_bit + ceil_log2(n) - 1;
	num_bits += (data.high_bit - data.low_bit + 1);
      } else {
	data.low_bit = -1;
      }
    }
  }
  return num_bits;
}


/* ====================================================================== */
/* DD variable access. */

/*
 * Returns the BDD variable for the given state variable.
 */
static DdNode* bdd_var(DdManager* dd_man, int i, bool primed = false) {
  return Cudd_bddIthVar(dd_man, 2*var_order[i] + (primed ? 1 : 0));
}


/*
 * Returns the ADD variable for the given state variable.
 */
static DdNode* add_var(DdManager* dd_man, int i, bool primed = false) {
  return Cudd_addIthVar(dd_man, 2*var_order[i] + (primed ? 1 : 0));
}


/* ====================================================================== */
/* formula_bdd */

/*
 * Constructs a BDD representing the given formula.
 */
static DdNode* formula_bdd(DdManager* dd_man, const StateFormula& formula,
			   bool primed = false) {
  if (formula.tautology() || formula.contradiction()) {
    /*
     * The formula is either TRUE or FALSE, so the BDD is either
     * constant 1 or 0.
     */
    DdNode* ddf = (formula.tautology() ?
		   Cudd_ReadOne(dd_man) : Cudd_ReadLogicZero(dd_man));
    Cudd_Ref(ddf);
    return ddf;
  }

  const Atom* af = dynamic_cast<const Atom*>(&formula);
  if (af != 0) {
    /*
     * The BDD for an atom is the `current-state' (or `next-state' if
     * primed is true) DD variable for the state variable represented
     * by the atom.
     */
    DdNode* ddf = bdd_var(dd_man, state_variables[af], primed);
    Cudd_Ref(ddf);
    return ddf;
  }

  const Negation* nf = dynamic_cast<const Negation*>(&formula);
  if (nf != 0) {
    /*
     * The BDD for a negation is the negation of the BDD for the
     * negand.
     */
    DdNode* ddn = formula_bdd(dd_man, nf->negand(), primed);
    DdNode* ddf = Cudd_Not(ddn);
    Cudd_Ref(ddf);
    Cudd_RecursiveDeref(dd_man, ddn);
    return ddf;
  }

  const Conjunction* cf = dynamic_cast<const Conjunction*>(&formula);
  if (cf != 0) {
    /*
     * The BDD for a conjunction is the conjunction of the BDDs for
     * the conjuncts.
     */
    DdNode* ddf = Cudd_ReadOne(dd_man);
    Cudd_Ref(ddf);
    for (FormulaList::const_reverse_iterator fi = cf->conjuncts().rbegin();
	 fi != cf->conjuncts().rend(); fi++) {
      DdNode* ddi = formula_bdd(dd_man, **fi, primed);
      DdNode* dda = Cudd_bddAnd(dd_man, ddf, ddi);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddf);
      Cudd_RecursiveDeref(dd_man, ddi);
      ddf = dda;
    }
    return ddf;
  }

  const Disjunction* df = dynamic_cast<const Disjunction*>(&formula);
  if (df != 0) {
    /*
     * The BDD for a disjunction is the disjunction of the BDDs for
     * the disjuncts.
     */
    DdNode* ddf = Cudd_ReadLogicZero(dd_man);
    Cudd_Ref(ddf);
    for (FormulaList::const_reverse_iterator fi = df->disjuncts().rbegin();
	 fi != df->disjuncts().rend(); fi++) {
      DdNode* ddi = formula_bdd(dd_man, **fi, primed);
      DdNode* ddo = Cudd_bddOr(dd_man, ddf, ddi);
      Cudd_Ref(ddo);
      Cudd_RecursiveDeref(dd_man, ddf);
      Cudd_RecursiveDeref(dd_man, ddi);
      ddf = ddo;
    }
    return ddf;
  }

  /*
   * No other types of formulae should appear in fully instantiated
   * event conditions and effect conditions.
   */
  throw std::logic_error("unexpected formula");
}


/* ====================================================================== */
/* phase_bdd */

/*
 * Constructs a BDD representing the given phase.
 */
static DdNode* phase_bdd(DdManager* dd_man, int ph,
			 int low_bit, int high_bit, bool primed = false) {
  DdNode* ddp = Cudd_ReadOne(dd_man);
  Cudd_Ref(ddp);
  if (low_bit >= 0) {
    for (int b = high_bit; b >= low_bit; b--) {
      int v = 1 << (high_bit - b);
      DdNode* ddb = bdd_var(dd_man, b, primed);
      if ((v & ph) == 0) {
	ddb = Cudd_Not(ddb);
      }
      DdNode* dda = Cudd_bddAnd(dd_man, ddb, ddp);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddp);
      ddp = dda;
    }
  }
  return ddp;
}


/* ====================================================================== */
/* state_bdd */

/*
 * Returns a BDD representing the given state.
 */
static DdNode* state_bdd(DdManager* dd_man, const AtomSet& atoms,
			 const std::vector<size_t>& phases) {
  /* This is going to be the BDD representing the given state. */
  DdNode* dds = Cudd_ReadOne(dd_man);
  Cudd_Ref(dds);

  /*
   * Set regular state variables to the values specified by the given
   * atom set.
   */
  for (std::map<int, const Atom*>::reverse_iterator ai =
	 dynamic_atoms.rbegin();
       ai != dynamic_atoms.rend(); ai++) {
    int i = (*ai).first;
    DdNode* ddv = bdd_var(dd_man, i);
    if (atoms.find((*ai).second) == atoms.end()) {
      ddv = Cudd_Not(ddv);
    }
    DdNode* ddt = Cudd_bddAnd(dd_man, ddv, dds);
    Cudd_Ref(ddt);
    Cudd_RecursiveDeref(dd_man, dds);
    dds = ddt;
  }

  /*
   * Set the values of phase variables.
   */
  int i = phases.size() - 1;
  for (std::map<const Event*, PHData>::reverse_iterator ei =
	 ph_events.rbegin();
       ei != ph_events.rend(); ei++, i--) {
    const PHData& data = (*ei).second;
    DdNode* ddi = phase_bdd(dd_man, phases[i], data.low_bit, data.high_bit);
    DdNode* ddt = Cudd_bddAnd(dd_man, ddi, dds);
    Cudd_Ref(ddt);
    Cudd_RecursiveDeref(dd_man, ddi);
    Cudd_RecursiveDeref(dd_man, dds);
    dds = ddt;
  }

  /* Return the BDD representing state. */
  return dds;
}


/* ====================================================================== */
/* initial_state_mtbdd */

/*
 * Returns a MTBDD representing the initial state distribution of the
 * given problem.
 */
static DdNode* initial_state_mtbdd(DdManager* dd_man, const Problem& problem) {
  AtomSet atoms;
  for (AtomSet::const_iterator ai = problem.init_atoms().begin();
       ai != problem.init_atoms().end(); ai++) {
    if (!PredicateTable::static_predicate((*ai)->predicate())) {
      atoms.insert(*ai);
    }
  }
  std::vector<size_t> phases(ph_events.size(), 0);
  DdNode* dds = state_bdd(dd_man, atoms, phases);
  DdNode* ddS = Cudd_BddToAdd(dd_man, dds);
  Cudd_Ref(ddS);
  Cudd_RecursiveDeref(dd_man, dds);
  // TODO: probabilistic initial effects
  return ddS;
}


/* ====================================================================== */
/* phase_identity_bdd */

/*
 * Constructs a BDD representing phase identity.
 */
static DdNode* phase_identity_bdd(DdManager* dd_man,
				  int low_bit, int high_bit) {
  DdNode* ddi;
  if (low_bit >= 0) {
    int m = high_bit - low_bit + 1;
    DdNode** xvars = new DdNode*[m];
    DdNode** yvars = new DdNode*[m];
    for (int b = high_bit; b >= low_bit; b--) {
      xvars[b - low_bit] = bdd_var(dd_man, b);
      yvars[b - low_bit] = bdd_var(dd_man, b, true);
    }
    ddi = Cudd_Xeqy(dd_man, m, xvars, yvars);
    Cudd_Ref(ddi);
    delete xvars;
    delete yvars;
  } else {
    ddi = Cudd_ReadOne(dd_man);
    Cudd_Ref(ddi);
  }
  return ddi;
}


/* ====================================================================== */
/* phase_update_bdd */

/*
 * Constructs a BDD representing phase updates.
 */
static DdNode* phase_update_bdd(DdManager* dd_man, const Event& event) {
  DdNode* ddu = Cudd_ReadOne(dd_man);
  Cudd_Ref(ddu);
  const std::map<const Event*, PHData>& phe = ph_events;
  for (std::map<const Event*, PHData>::const_reverse_iterator ei =
	 phe.rbegin();
       ei != phe.rend(); ei++) {
    const Event& e = *(*ei).first;
    if (&e != &event && (event.controllable() || !e.controllable())) {
      const PHData& data = (*ei).second;
      DdNode* ddo;
      if (event.controllable() && e.controllable()) {
	ddo = phase_bdd(dd_man, 0, data.low_bit, data.high_bit, true);
      } else {
	DdNode* ddi = phase_identity_bdd(dd_man, data.low_bit, data.high_bit);
	DdNode* ddc = formula_bdd(dd_man, e.condition(), true);
	DdNode* dda = Cudd_bddAnd(dd_man, ddi, ddc);
	Cudd_Ref(dda);
	Cudd_RecursiveDeref(dd_man, ddi);
	DdNode* ddn = Cudd_Not(ddc);
	Cudd_Ref(ddn);
	Cudd_RecursiveDeref(dd_man, ddc);
	DdNode* dd0 = phase_bdd(dd_man, 0, data.low_bit, data.high_bit, true);
	DdNode* ddb = Cudd_bddAnd(dd_man, dd0, ddn);
	Cudd_Ref(ddb);
	Cudd_RecursiveDeref(dd_man, dd0);
	Cudd_RecursiveDeref(dd_man, ddn);
	ddo = Cudd_bddOr(dd_man, dda, ddb);
	Cudd_Ref(ddo);
	Cudd_RecursiveDeref(dd_man, dda);
	Cudd_RecursiveDeref(dd_man, ddb);
      }
      DdNode* dda = Cudd_bddAnd(dd_man, ddo, ddu);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddo);
      Cudd_RecursiveDeref(dd_man, ddu);
      ddu = dda;
    }
  }
  return ddu;
}


/* ====================================================================== */
/* action_phase_update_bdd */

/*
 * Constructs a BDD representing action phase updates.
 */
static DdNode* action_phase_update_bdd(DdManager* dd_man,
				       const Event* action) {
  DdNode* ddu = Cudd_ReadOne(dd_man);
  Cudd_Ref(ddu);
  const std::map<const Event*, PHData>& phe = ph_events;
  for (std::map<const Event*, PHData>::const_reverse_iterator ei =
	 phe.rbegin();
       ei != phe.rend(); ei++) {
    const Event& e = *(*ei).first;
    if (e.controllable()) {
      const PHData& data = (*ei).second;
      DdNode* ddo;
      if (action != &e) {
	ddo = phase_bdd(dd_man, 0, data.low_bit, data.high_bit, true);
      } else {
	DdNode* ddi = phase_identity_bdd(dd_man, data.low_bit, data.high_bit);
	DdNode* ddc = formula_bdd(dd_man, e.condition(), true);
	DdNode* dda = Cudd_bddAnd(dd_man, ddi, ddc);
	Cudd_Ref(dda);
	Cudd_RecursiveDeref(dd_man, ddi);
	DdNode* ddn = Cudd_Not(ddc);
	Cudd_Ref(ddn);
	Cudd_RecursiveDeref(dd_man, ddc);
	DdNode* dd0 = phase_bdd(dd_man, 0, data.low_bit, data.high_bit, true);
	DdNode* ddb = Cudd_bddAnd(dd_man, dd0, ddn);
	Cudd_Ref(ddb);
	Cudd_RecursiveDeref(dd_man, dd0);
	Cudd_RecursiveDeref(dd_man, ddn);
	ddo = Cudd_bddOr(dd_man, dda, ddb);
	Cudd_Ref(ddo);
	Cudd_RecursiveDeref(dd_man, dda);
	Cudd_RecursiveDeref(dd_man, ddb);
      }
      DdNode* dda = Cudd_bddAnd(dd_man, ddo, ddu);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddo);
      Cudd_RecursiveDeref(dd_man, ddu);
      ddu = dda;
    }
  }
  return ddu;
}


/* ====================================================================== */
/* phase_filter_mtbdd */

/*
 * Constructs an MTBDD representing possible phase configurations.
 */
static DdNode* phase_filter_mtbdd(DdManager* dd_man, bool primed = false) {
  DdNode* ddf = Cudd_ReadOne(dd_man);
  Cudd_Ref(ddf);
  const std::map<const Event*, PHData>& phe = ph_events;
  for (std::map<const Event*, PHData>::const_reverse_iterator ei =
	 phe.rbegin();
       ei != phe.rend(); ei++) {
    const Event& event = *(*ei).first;
    const PHData& data = (*ei).second;
    DdNode* ddv = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddv);
    if (data.low_bit >= 0) {
      for (int b = data.high_bit; b >= data.low_bit; b--) {
	DdNode* ddk = Cudd_addConst(dd_man, 1 << (data.high_bit - b));
	Cudd_Ref(ddk);
	DdNode* ddb = add_var(dd_man, b, primed);
	Cudd_Ref(ddb);
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddk, ddb);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddk);
	Cudd_RecursiveDeref(dd_man, ddb);
	ddb = Cudd_addApply(dd_man, Cudd_addPlus, ddm, ddv);
	Cudd_Ref(ddb);
	Cudd_RecursiveDeref(dd_man, ddm);
	Cudd_RecursiveDeref(dd_man, ddv);
	ddv = ddb;
      }
    }
    DdNode* ddi = Cudd_addBddInterval(dd_man, ddv, 0, data.params.n - 1);
    Cudd_Ref(ddi);
    Cudd_RecursiveDeref(dd_man, ddv);
    DdNode* ddc = formula_bdd(dd_man, event.condition(), primed);
    DdNode* dda = Cudd_bddAnd(dd_man, ddi, ddc);
    Cudd_Ref(dda);
    Cudd_RecursiveDeref(dd_man, ddi);
    DdNode* dd0 = phase_bdd(dd_man, 0, data.low_bit, data.high_bit, primed);
    DdNode* ddn = Cudd_Not(ddc);
    Cudd_Ref(ddn);
    Cudd_RecursiveDeref(dd_man, ddc);
    DdNode* ddb = Cudd_bddAnd(dd_man, dd0, ddn);
    Cudd_Ref(ddb);
    Cudd_RecursiveDeref(dd_man, dd0);
    Cudd_RecursiveDeref(dd_man, ddn);
    DdNode* ddo = Cudd_bddOr(dd_man, dda, ddb);
    Cudd_Ref(ddo);
    Cudd_RecursiveDeref(dd_man, dda);
    Cudd_RecursiveDeref(dd_man, ddb);
    dda = Cudd_bddAnd(dd_man, ddo, ddf);
    Cudd_Ref(dda);
    Cudd_RecursiveDeref(dd_man, ddo);
    Cudd_RecursiveDeref(dd_man, ddf);
    ddf = dda;
  }
  DdNode* ddF = Cudd_BddToAdd(dd_man, ddf);
  Cudd_Ref(ddF);
  Cudd_RecursiveDeref(dd_man, ddf);
  return ddF;
}


/* ====================================================================== */
/* state_identity_bdd */

/*
 * Constructs a BDD representing state variable identity.
 */
static DdNode* state_identity_bdd(DdManager* dd_man) {
  int n = nvars - aux_vars - ph_vars;
  DdNode** xvars = new DdNode*[n];
  DdNode** yvars = new DdNode*[n];
  for (int i = 0; i < n; i++) {
    int v = i + aux_vars + ph_vars;
    xvars[i] = bdd_var(dd_man, v);
    yvars[i] = bdd_var(dd_man, v, true);
  }
  DdNode* ddi = Cudd_Xeqy(dd_man, n, xvars, yvars);
  Cudd_Ref(ddi);
  delete xvars;
  delete yvars;
  return ddi;
}


/* ====================================================================== */
/* reward_rate_mtbdd */

/*
 * Constructs an MTBDD representing the reward rate vector for the
 * given reward rates.
 */
static DdNode* reward_rate_mtbdd(DdManager* dd_man,
				 const RewardRateList& reward_rates,
				 DdNode* ddng) {
  /* This is going to be the MTBDD representing the reward rate vector. */
  DdNode* ddR = Cudd_ReadZero(dd_man);
  Cudd_Ref(ddR);

  /*
   * Add the specified reward rates.
   */
  for (RewardRateList::const_reverse_iterator ri = reward_rates.rbegin();
       ri != reward_rates.rend(); ri++) {
    DdNode* ddt = formula_bdd(dd_man, *(*ri).first);
    DdNode* dda = Cudd_bddAnd(dd_man, ddt, ddng);
    Cudd_Ref(dda);
    Cudd_RecursiveDeref(dd_man, ddt);
    DdNode* ddf = Cudd_BddToAdd(dd_man, dda);
    Cudd_Ref(ddf);
    Cudd_RecursiveDeref(dd_man, dda);
    DdNode* ddr = Cudd_addConst(dd_man, (*ri).second.double_value());
    Cudd_Ref(ddr);
    ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddf, ddr);
    Cudd_Ref(ddt);
    Cudd_RecursiveDeref(dd_man, ddf);
    Cudd_RecursiveDeref(dd_man, ddr);
    ddr = Cudd_addApply(dd_man, Cudd_addPlus, ddt, ddR);
    Cudd_Ref(ddr);
    Cudd_RecursiveDeref(dd_man, ddt);
    Cudd_RecursiveDeref(dd_man, ddR);
    ddR = ddr;
  }

  /* Return the MTBDD representing the reward rate vector. */
  return ddR;
}


/* ====================================================================== */
/* goal_reward_mtbdd */

/*
 * Constructs an MTBDD representing goal rewards.
 */
static DdNode* goal_reward_mtbdd(DdManager* dd_man, const Rational& reward,
				 DdNode* ddng, DdNode* ddgp) {
  DdNode* ddgr;
  if (reward == 0) {
    ddgr = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddgr);
  } else {
    DdNode* ddr = Cudd_addConst(dd_man, reward.double_value());
    Cudd_Ref(ddr);
    DdNode* dda = Cudd_bddAnd(dd_man, ddng, ddgp);
    Cudd_Ref(dda);
    DdNode* ddt = Cudd_BddToAdd(dd_man, dda);
    Cudd_Ref(ddt);
    Cudd_RecursiveDeref(dd_man, dda);
    ddgr = Cudd_addApply(dd_man, Cudd_addTimes, ddt, ddr);
    Cudd_Ref(ddgr);
    Cudd_RecursiveDeref(dd_man, ddt);
    Cudd_RecursiveDeref(dd_man, ddr);
  }
  return ddgr;
}


/* ====================================================================== */
/* holding_time_mtbdds */

/*
 * Computes the holding time MTBDDs for the given events.
 */
static void holding_time_mtbdds(DdManager* dd_man,
				const EventSet& events, double alpha) {
  DdNode* ddL = Cudd_addConst(dd_man, alpha);
  Cudd_Ref(ddL);
  for (EventSet::const_iterator ei = events.begin();
       ei != events.end(); ei++) {
    const Event& event = **ei;
    DdNode* ddLa;
    DdNode* ddc = formula_bdd(dd_man, event.condition());
    const Exponential* exp_dist =
      dynamic_cast<const Exponential*>(&event.delay());
    if (exp_dist != 0) {
      DdNode* ddC = Cudd_BddToAdd(dd_man, ddc);
      Cudd_Ref(ddC);
      DdNode* ddl = Cudd_addConst(dd_man, exp_dist->rate());
      Cudd_Ref(ddl);
      ddLa = Cudd_addApply(dd_man, Cudd_addTimes, ddC, ddl);
      Cudd_Ref(ddLa);
      Cudd_RecursiveDeref(dd_man, ddC);
      Cudd_RecursiveDeref(dd_man, ddl);
    } else {
      ddLa = Cudd_ReadZero(dd_man);
      Cudd_Ref(ddLa);
      std::map<const Event*, PHData>::const_iterator phi =
	ph_events.find(&event);
      if (phi != ph_events.end()) {
	const PHParameters& params = (*phi).second.params;
	for (TransitionRateMap::const_iterator ri = params.rates.begin();
	     ri != params.rates.end(); ri++) {
	  DdNode* ddp = phase_bdd(dd_man, (*ri).first.first,
				  (*phi).second.low_bit,
				  (*phi).second.high_bit);
	  DdNode* dda = Cudd_bddAnd(dd_man, ddp, ddc);
	  Cudd_Ref(dda);
	  Cudd_RecursiveDeref(dd_man, ddp);
	  DdNode* ddC = Cudd_BddToAdd(dd_man, dda);
	  Cudd_Ref(ddC);
	  Cudd_RecursiveDeref(dd_man, dda);
	  DdNode* ddl = Cudd_addConst(dd_man, (*ri).second);
	  Cudd_Ref(ddl);
	  DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddC, ddl);
	  Cudd_Ref(ddm);
	  Cudd_RecursiveDeref(dd_man, ddC);
	  Cudd_RecursiveDeref(dd_man, ddl);
	  DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddm, ddLa);
	  Cudd_Ref(dds);
	  Cudd_RecursiveDeref(dd_man, ddm);
	  Cudd_RecursiveDeref(dd_man, ddLa);
	  ddLa = dds;
	}
      } else {
	throw std::logic_error("no phase data for general distribution");
      }
    }
    Cudd_RecursiveDeref(dd_man, ddc);
    if (event.controllable()) {
      holding_times.insert(std::make_pair(&event, ddLa));
    } else {
      DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddLa, ddL);
      Cudd_Ref(dds);
      Cudd_RecursiveDeref(dd_man, ddLa);
      Cudd_RecursiveDeref(dd_man, ddL);
      ddL = dds;
    }
  }
  DdNode* dd1 = Cudd_ReadOne(dd_man);
  Cudd_Ref(dd1);
  DdNode* ddH = Cudd_addApply(dd_man, Cudd_addDivide, dd1, ddL);
  Cudd_Ref(ddH);
  if (verbosity > 2) {
    std::cout << std::endl << "holding times for <idle>:" << std::endl;
    Cudd_PrintDebug(dd_man, ddH, 2*nvars, 2);
  }
  for (std::map<const Event*, DdNode*>::iterator ei = holding_times.begin();
       ei != holding_times.end(); ei++) {
    DdNode* ddLa = Cudd_addApply(dd_man, Cudd_addPlus, ddL, (*ei).second);
    Cudd_Ref(ddLa);
    Cudd_RecursiveDeref(dd_man, (*ei).second);
    (*ei).second = Cudd_addApply(dd_man, Cudd_addDivide, dd1, ddLa);
    Cudd_Ref((*ei).second);
    Cudd_RecursiveDeref(dd_man, ddLa);
    if (verbosity > 2) {
      std::cout << std::endl << "holding times for " << *(*ei).first << ':'
		<< std::endl;
      Cudd_PrintDebug(dd_man, (*ei).second, 2*nvars, 2);
    }
  }
  Cudd_RecursiveDeref(dd_man, ddL);
  Cudd_RecursiveDeref(dd_man, dd1);
  holding_times.insert(std::pair<const Event*, DdNode*>(0, ddH));
}


/* ====================================================================== */
/* effect_dbn */

/*
 * Combines two CPTs.
 */
static void combine_cpts(std::map<int, DdNode*>& cpts1,
			 const std::map<int, DdNode*>& cpts2,
			 DdManager* dd_man) {
  std::map<int, DdNode*>::iterator ci = cpts1.begin();
  std::map<int, DdNode*>::const_iterator cj = cpts2.begin();
  while (ci != cpts1.end() && cj != cpts2.end()) {
    int i = (*ci).first;
    if (i < (*cj).first) {
      ci++;
    } else if ((*cj).first < i) {
      cpts1.insert(*cj);
      ci = cpts1.find(i);
      cj++;
    } else {
      DdNode* ddo = Cudd_bddOr(dd_man, (*ci).second, (*cj).second);
      Cudd_Ref(ddo);
      Cudd_RecursiveDeref(dd_man, (*ci).second);
      Cudd_RecursiveDeref(dd_man, (*cj).second);
      (*ci).second = ddo;
      ci++;
      cj++;
    }
  }
  while (cj != cpts2.end()) {
    cpts1.insert(*cj);
    cj++;
  }
}


/*
 * Constructs a DBN representing the given effect.
 */
static DdNode* effect_dbn(std::map<int, DdNode*>& cpts,
			  std::map<int, DdNode*>& ncpts,
			  std::vector<DdNode*>& acpts, int& next_aux,
			  DdManager* dd_man, DdNode* condition_bdd,
			  DdNode* aux_cond, const Effect& effect) {
  if (condition_bdd == Cudd_ReadLogicZero(dd_man) || effect.empty()) {
    /*
     * The condition is false or the effect is empty.
     */
    DdNode* ddR = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddR);
    return ddR;
  }

  const UpdateEffect* ue = dynamic_cast<const UpdateEffect*>(&effect);
  if (ue != 0) {
    /*
     * Only reward updates are supported.
     */
    const Fluent& fluent = ue->update().fluent();
    if (reward_function == 0 || fluent.function() != *reward_function) {
      throw std::logic_error("numeric state variables not supported");
    }
    ValueMap values;
    values[&fluent] = 0;
    ue->update().affect(values);
    DdNode* ddc = Cudd_BddToAdd(dd_man, condition_bdd);
    Cudd_Ref(ddc);
    DdNode* ddr = Cudd_addConst(dd_man, values[&fluent].double_value());
    Cudd_Ref(ddr);
    DdNode* ddR = Cudd_addApply(dd_man, Cudd_addTimes, ddc, ddr);
    Cudd_Ref(ddR);
    Cudd_RecursiveDeref(dd_man, ddc);
    Cudd_RecursiveDeref(dd_man, ddr);
    return ddR;
  }

  const SimpleEffect* se = dynamic_cast<const SimpleEffect*>(&effect);
  if (se != 0) {
    /*
     * Set the CPT for the affected variable, taking the condition
     * into account.
     */
    bool is_true = typeid(*se) == typeid(AddEffect);
    int v = state_variables[&se->atom()];
    Cudd_Ref(aux_cond);
    if (is_true) {
      cpts.insert(std::make_pair(v, aux_cond));
    } else {
      ncpts.insert(std::make_pair(v, aux_cond));
    }
    DdNode* ddR = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddR);
    return ddR;
  }

  const ConjunctiveEffect* ce =
    dynamic_cast<const ConjunctiveEffect*>(&effect);
  if (ce != 0) {
    /*
     * Take the maximum of the CPTs for the conjuncts.
     */
    EffectList::const_iterator ei = ce->conjuncts().begin();
    DdNode* ddR = effect_dbn(cpts, ncpts, acpts, next_aux,
			     dd_man, condition_bdd, aux_cond, **ei);
    for (ei++; ei != ce->conjuncts().end(); ei++) {
      std::map<int, DdNode*> c_cpts, c_ncpts;
      DdNode* ddr = effect_dbn(c_cpts, c_ncpts, acpts, next_aux,
			       dd_man, condition_bdd, aux_cond, **ei);
      combine_cpts(cpts, c_cpts, dd_man);
      combine_cpts(ncpts, c_ncpts, dd_man);
      DdNode* dda = Cudd_addApply(dd_man, Cudd_addPlus, ddR, ddr);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddR);
      Cudd_RecursiveDeref(dd_man, ddr);
      ddR = dda;
    }
    return ddR;
  }

  const ConditionalEffect* we =
    dynamic_cast<const ConditionalEffect*>(&effect);
  if (we != 0) {
    /*
     * Add to condition and recurse.
     */
    DdNode* ddf = formula_bdd(dd_man, we->condition());
    DdNode* ddc = Cudd_bddAnd(dd_man, condition_bdd, ddf);
    Cudd_Ref(ddc);
    DdNode* dda = Cudd_bddAnd(dd_man, aux_cond, ddf);
    Cudd_Ref(dda);
    Cudd_RecursiveDeref(dd_man, ddf);
    DdNode* ddR = effect_dbn(cpts, ncpts, acpts, next_aux,
			     dd_man, ddc, dda, we->effect());
    Cudd_RecursiveDeref(dd_man, ddc);
    Cudd_RecursiveDeref(dd_man, dda);
    return ddR;
  }

  const ProbabilisticEffect* pe =
    dynamic_cast<const ProbabilisticEffect*>(&effect);
  if (pe != 0) {
    /*
     * Handle probabilistic effect by introducing auxiliary state
     * variables.
     */
    Rational p_rest = 1;
    int n = pe->size();
    for (int i = 0; i < n; i++) {
      p_rest = p_rest - pe->probability(i);
    }
    if (p_rest > 0) {
      n++;
    }
    int low_bit = next_aux;
    int high_bit = low_bit + ceil_log2(n) - 1;
    next_aux = high_bit + 1;
    DdNode* ddP = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddP);
    size_t a_pos = acpts.size();
    acpts.push_back(ddP);
    DdNode* ddR = Cudd_ReadZero(dd_man);
    Cudd_Ref(ddR);
    for (int i = 0; i < n; i++) {
      Rational p = (i + 1 == n && p_rest > 0) ? p_rest : pe->probability(i);
      DdNode* ddi = Cudd_ReadOne(dd_man);
      Cudd_Ref(ddi);
      for (int b = high_bit; b >= low_bit; b--) {
	int v = 1 << (high_bit - b);
	DdNode* ddb = bdd_var(dd_man, b, true);
	if ((v & i) == 0) {
	  ddb = Cudd_Not(ddb);
	}
	DdNode* dda = Cudd_bddAnd(dd_man, ddb, ddi);
	Cudd_Ref(dda);
	Cudd_RecursiveDeref(dd_man, ddi);
	ddi = dda;
      }
      DdNode* ddc = Cudd_bddAnd(dd_man, aux_cond, ddi);
      Cudd_Ref(ddc);
      DdNode* ddI = Cudd_BddToAdd(dd_man, ddi);
      Cudd_Ref(ddI);
      Cudd_RecursiveDeref(dd_man, ddi);
      DdNode* ddp = Cudd_addConst(dd_man, p.double_value());
      Cudd_Ref(ddp);
      ddi = Cudd_addApply(dd_man, Cudd_addTimes, ddI, ddp);
      Cudd_Ref(ddi);
      Cudd_RecursiveDeref(dd_man, ddI);
      DdNode* ddt = Cudd_addApply(dd_man, Cudd_addPlus, ddi, ddP);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, ddi);
      Cudd_RecursiveDeref(dd_man, ddP);
      ddP = ddt;
      if (i + 1 < n || p_rest == 0) {
	DdNode* ddr;
	if (i == 0) {
	  ddr = effect_dbn(cpts, ncpts, acpts, next_aux,
			   dd_man, condition_bdd, ddc, pe->effect(i));
	} else {
	  std::map<int, DdNode*> o_cpts, o_ncpts;
	  ddr = effect_dbn(o_cpts, o_ncpts, acpts, next_aux,
			   dd_man, condition_bdd, ddc, pe->effect(i));
	  combine_cpts(cpts, o_cpts, dd_man);
	  combine_cpts(ncpts, o_ncpts, dd_man);
	}
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddr, ddp);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddr);
	ddr = Cudd_addApply(dd_man, Cudd_addPlus, ddR, ddm);
	Cudd_Ref(ddr);
	Cudd_RecursiveDeref(dd_man, ddR);
	Cudd_RecursiveDeref(dd_man, ddm);
	ddR = ddr;
      }
      Cudd_RecursiveDeref(dd_man, ddc);
      Cudd_RecursiveDeref(dd_man, ddp);
    }
    acpts[a_pos] = ddP;
    return ddR;
  }

  /*
   * No other types of effects exist.
   */
  throw std::logic_error("unexpected effect");
}


/* ====================================================================== */
/* event_dbn */

/*
 * Constructs a DBN for the given event.
 */
static DdNode* event_dbn(std::map<int, DdNode*>& cpts,
			 std::vector<DdNode*>& acpts, int& next_aux,
			 DdManager* dd_man, const Event& event) {
  std::map<int, DdNode*> ncpts;
  DdNode* ddc = Cudd_ReadOne(dd_man);
  Cudd_Ref(ddc);
  DdNode* ddR = effect_dbn(cpts, ncpts, acpts, next_aux, dd_man,
			   ddc, ddc, event.effect());
  Cudd_RecursiveDeref(dd_man, ddc);
  for (int i = aux_vars + ph_vars; i < nvars; i++) {
    std::map<int, DdNode*>::iterator ci = cpts.find(i);
    std::map<int, DdNode*>::const_iterator cj = ncpts.find(i);
    if (ci != cpts.end()) {
      if (cj != ncpts.end()) {
	DdNode* dda = Cudd_bddAnd(dd_man, (*ci).second, (*cj).second);
	Cudd_Ref(dda);
	if (dda != Cudd_ReadLogicZero(dd_man)) {
	  throw std::logic_error("event `" + event.name()
				 + "' has inconsistent effects");
	}
	Cudd_RecursiveDeref(dd_man, dda);
	DdNode* ddv = Cudd_Not(bdd_var(dd_man, i));
	DdNode* ddo = Cudd_bddOr(dd_man, ddv, (*cj).second);
	Cudd_Ref(ddo);
	Cudd_RecursiveDeref(dd_man, (*cj).second);
	DdNode* ddn = Cudd_Not(ddo);
	Cudd_Ref(ddn);
	Cudd_RecursiveDeref(dd_man, ddo);
	ddo = Cudd_bddOr(dd_man, ddn, (*ci).second);
	Cudd_Ref(ddo);
	Cudd_RecursiveDeref(dd_man, ddn);
	Cudd_RecursiveDeref(dd_man, (*ci).second);
	(*ci).second = ddo;
      } else {
	DdNode* ddv = bdd_var(dd_man, i);
	DdNode* ddo = Cudd_bddOr(dd_man, ddv, (*ci).second);
	Cudd_Ref(ddo);
	Cudd_RecursiveDeref(dd_man, (*ci).second);
	(*ci).second = ddo;
      }
    } else if (cj != ncpts.end()) {
      DdNode* ddv = Cudd_Not(bdd_var(dd_man, i));
      DdNode* ddo = Cudd_bddOr(dd_man, ddv, (*cj).second);
      Cudd_Ref(ddo);
      Cudd_RecursiveDeref(dd_man, (*cj).second);
      DdNode* ddn = Cudd_Not(ddo);
      Cudd_Ref(ddn);
      Cudd_RecursiveDeref(dd_man, ddo);
      cpts.insert(std::make_pair(i, ddn));
    } else {
      DdNode* ddv = bdd_var(dd_man, i);
      Cudd_Ref(ddv);
      cpts.insert(std::make_pair(i, ddv));
    }
  }
  return ddR;
}


/* ====================================================================== */
/* event_mtbdds */

/*
 * Constructs MTBDDs representing transition probability matrices and
 * state reward vectors for the given events.
 */
static void event_mtbdds(DdManager* dd_man, const EventSet& events,
			 DdNode* ddng, DdNode* ddgr, DdNode* reward_rates) {
  /*
   * Compute state variable cube.
   */
  DdNode* sv_cube;
  if (ddgr != Cudd_ReadZero(dd_man)) {
    DdNode** col_variables = new DdNode*[nvars - aux_vars - ph_vars];
    for (int i = aux_vars + ph_vars; i < nvars; i++) {
      col_variables[i - aux_vars - ph_vars] = add_var(dd_man, i, true);
      Cudd_Ref(col_variables[i - aux_vars - ph_vars]);
    }
    sv_cube = Cudd_addComputeCube(dd_man, col_variables, 0,
				  nvars - aux_vars - ph_vars);
    Cudd_Ref(sv_cube);
    for (int i = aux_vars + ph_vars; i < nvars; i++) {
      Cudd_RecursiveDeref(dd_man, col_variables[i - aux_vars - ph_vars]);
    }
    delete col_variables;
  } else {
    sv_cube = 0;
  }

  /*
   * Compute BDD representing state variable identity.
   */
  DdNode* state_id = state_identity_bdd(dd_man);

  /*
   * Compute BDD representing action phase updates.
   */
  DdNode* ddU = action_phase_update_bdd(dd_man, 0);
  DdNode* ddUi = Cudd_BddToAdd(dd_man, ddU);
  Cudd_Ref(ddUi);
  Cudd_RecursiveDeref(dd_man, ddU);
  action_phase_updates.insert(std::pair<const Event*, DdNode*>(0, ddUi));
  if (verbosity > 2) {
    std::cout << std::endl << "Phase update for <idle>:" << std::endl;
    Cudd_PrintDebug(dd_man, ddUi, 2*nvars, 2);
  }

  /*
   * Process each event.
   */
  DdNode* dd0 = Cudd_ReadZero(dd_man);
  Cudd_Ref(dd0);
  action_rewards.insert(std::pair<const Event*, DdNode*>(0, dd0));
  for (EventSet::const_iterator ei = events.begin();
       ei != events.end(); ei++) {
    const Event& event = **ei;
    if (verbosity > 2) {
      std::cout << std::endl << "processing event " << event << "..."
		<< std::endl;
    }

    /*
     * Generate DBN for event effect.
     */
    std::map<int, DdNode*> cpts;
    std::vector<DdNode*> acpts;
    int next_aux = 0;
    DdNode* ddR = event_dbn(cpts, acpts, next_aux, dd_man, event);
    DdNode* ddP = Cudd_ReadOne(dd_man);
    Cudd_Ref(ddP);
    if (verbosity > 3) {
      std::cout << std::endl << "DBN for " << event << ':' << std::endl;
    }

    /*
     * Compute transition probability matrix from DBN.
     */
    const std::vector<int>& ovars = ordered_vars;
    for (std::vector<int>::const_reverse_iterator oi = ovars.rbegin();
	 oi != ovars.rend(); oi++) {
      std::map<int, DdNode*>::const_iterator ci = cpts.find(*oi);
      if (ci != cpts.end()) {
	if (verbosity > 3) {
	  std::cout << "CPT for " << *dynamic_atoms[(*ci).first] << ':'
		    << std::endl;
	  Cudd_PrintDebug(dd_man, (*ci).second, 2*nvars, 2);
	}
	DdNode* ddi = (*ci).second;
	ddi = Cudd_BddToAdd(dd_man, ddi);
	Cudd_Ref(ddi);
	DdNode* ddv = add_var(dd_man, (*ci).first, true);
	Cudd_Ref(ddv);
	DdNode* ddnv = Cudd_addCmpl(dd_man, ddv);
	Cudd_Ref(ddnv);
	DdNode* ddp = Cudd_addApply(dd_man, Cudd_addTimes, ddv, ddi);
	Cudd_Ref(ddp);
	Cudd_RecursiveDeref(dd_man, ddv);
	DdNode* dd1 = Cudd_ReadOne(dd_man);
	Cudd_Ref(dd1);
	DdNode* ddn = Cudd_addApply(dd_man, Cudd_addMinus, dd1, ddi);
	Cudd_Ref(ddn);
	Cudd_RecursiveDeref(dd_man, dd1);
	Cudd_RecursiveDeref(dd_man, ddi);
	Cudd_RecursiveDeref(dd_man, (*ci).second);
	DdNode* ddq = Cudd_addApply(dd_man, Cudd_addTimes, ddnv, ddn);
	Cudd_Ref(ddq);
	Cudd_RecursiveDeref(dd_man, ddnv);
	Cudd_RecursiveDeref(dd_man, ddn);
	DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddp, ddq);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddp);
	Cudd_RecursiveDeref(dd_man, ddq);
	ddp = Cudd_addApply(dd_man, Cudd_addTimes, dds, ddP);
	Cudd_Ref(ddp);
	Cudd_RecursiveDeref(dd_man, dds);
	Cudd_RecursiveDeref(dd_man, ddP);
	ddP = ddp;
      }
    }
    const std::vector<DdNode*>& acpts_c = acpts;
    for (std::vector<DdNode*>::const_reverse_iterator ai = acpts_c.rbegin();
	 ai != acpts_c.rend(); ai++) {
      DdNode* ddA = *ai;
      if (verbosity > 3) {
	std::cout << "CPT for auxiliary variables:" << std::endl;
	Cudd_PrintDebug(dd_man, ddA, 2*nvars, 2);
      }
      DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddA, ddP);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddA);
      Cudd_RecursiveDeref(dd_man, ddP);
      ddP = ddm;
    }
    if (next_aux > 0) {
      DdNode** aux_variables = new DdNode*[next_aux];
      for (int i = 0; i < next_aux; i++) {
	aux_variables[i] = add_var(dd_man, i, true);
	Cudd_Ref(aux_variables[i]);
      }
      DdNode* aux_cube = Cudd_addComputeCube(dd_man, aux_variables, 0,
					     next_aux);
      Cudd_Ref(aux_cube);
      for (int i = 0; i < next_aux; i++) {
	Cudd_RecursiveDeref(dd_man, aux_variables[i]);
      }
      delete aux_variables;
      DdNode* ddm = Cudd_addExistAbstract(dd_man, ddP, aux_cube);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddP);
      Cudd_RecursiveDeref(dd_man, aux_cube);
      ddP = ddm;
    }
    if (verbosity > 3) {
      std::cout << std::endl << "Probability matrix for " << event << ':'
		<< std::endl;
      Cudd_PrintDebug(dd_man, ddP, 2*nvars, 2);
    }

    /*
     * Add goal reward to reward vector.
     */
    if (sv_cube != 0) {
      DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddgr, ddP);
      Cudd_Ref(ddt);
      DdNode* dde = Cudd_addExistAbstract(dd_man, ddt, sv_cube);
      Cudd_Ref(dde);
      Cudd_RecursiveDeref(dd_man, ddt);
      ddt = Cudd_addApply(dd_man, Cudd_addPlus, dde, ddR);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, dde);
      Cudd_RecursiveDeref(dd_man, ddR);
      ddR = ddt;
    }
    if (verbosity > 3) {
      std::cout << std::endl << "Reward vector for " << event << ':'
		<< std::endl;
      Cudd_PrintDebug(dd_man, ddR, 2*nvars, 2);
    }

    /*
     * Generate transition probability matrices for each phase transition.
     */
    DdNode* ddf = formula_bdd(dd_man, event.condition());
    DdNode* ddc = Cudd_bddAnd(dd_man, ddf, ddng);
    Cudd_Ref(ddc);
    Cudd_RecursiveDeref(dd_man, ddf);
    const Exponential* exp_dist =
      dynamic_cast<const Exponential*>(&event.delay());
    if (exp_dist != 0) {
      DdNode* ddu = phase_update_bdd(dd_man, event);
      DdNode* dda = Cudd_bddAnd(dd_man, ddc, ddu);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddu);
      DdNode* ddA = Cudd_BddToAdd(dd_man, dda);
      Cudd_Ref(ddA);
      Cudd_RecursiveDeref(dd_man, dda);
      DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddA, ddP);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddA);
      DdNode* ddl = Cudd_addConst(dd_man, exp_dist->rate());
      Cudd_Ref(ddl);
      ddA = Cudd_addApply(dd_man, Cudd_addTimes, ddl, ddm);
      Cudd_Ref(ddA);
      Cudd_RecursiveDeref(dd_man, ddm);
      DdNode* ddC = Cudd_BddToAdd(dd_man, ddc);
      Cudd_Ref(ddC);
      ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddC, ddR);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddC);
      DdNode* ddRe = Cudd_addApply(dd_man, Cudd_addTimes, ddl, ddm);
      Cudd_Ref(ddRe);
      Cudd_RecursiveDeref(dd_man, ddl);
      Cudd_RecursiveDeref(dd_man, ddm);
      if (verbosity > 2) {
	std::cout << std::endl << "Transition matrix for " << event << ':'
		  << std::endl;
	Cudd_PrintDebug(dd_man, ddA, 2*nvars, 2);
	std::cout << std::endl << "Reward vector for " << event << ':'
		  << std::endl;
	Cudd_PrintDebug(dd_man, ddRe, 2*nvars, 2);
      }
      if (event.controllable()) {
	action_transitions.insert(std::make_pair(&event, ddA));
	action_rewards.insert(std::make_pair(&event, ddRe));
      } else {
	event_transitions.insert(std::make_pair(&event, ddA));
	DdNode*& ddr = action_rewards[0];
	DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddRe, ddr);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddRe);
	Cudd_RecursiveDeref(dd_man, ddr);
	ddr = dds;
      }
    } else {
      DdNode* ddRe = Cudd_ReadZero(dd_man);
      Cudd_Ref(ddRe);
      std::map<const Event*, PHData>::const_iterator phi =
	ph_events.find(&event);
      if (phi != ph_events.end()) {
	const PHParameters& params = (*phi).second.params;
	int low_bit = (*phi).second.low_bit;
	int high_bit = (*phi).second.high_bit;
	for (TransitionRateMap::const_iterator ri = params.rates.begin();
	     ri != params.rates.end(); ri++) {
	  size_t from = (*ri).first.first;
	  size_t to = (*ri).first.second;
	  DdNode* ph1 = phase_bdd(dd_man, from, low_bit, high_bit);
	  DdNode* ph2 = phase_bdd(dd_man, to % params.n,
				  low_bit, high_bit, true);
	  DdNode* dda = Cudd_bddAnd(dd_man, ph1, ddc);
	  Cudd_Ref(dda);
	  Cudd_RecursiveDeref(dd_man, ph1);
	  DdNode* ddC = Cudd_BddToAdd(dd_man, dda);
	  Cudd_Ref(ddC);
	  DdNode* ddt = Cudd_bddAnd(dd_man, dda, ph2);
	  Cudd_Ref(ddt);
	  Cudd_RecursiveDeref(dd_man, dda);
	  Cudd_RecursiveDeref(dd_man, ph2);
	  if (to < params.n) {
	    dda = Cudd_bddAnd(dd_man, ddt, state_id);
	    Cudd_Ref(dda);
	    Cudd_RecursiveDeref(dd_man, ddt);
	    ddt = dda;
	  }
	  DdNode* ddu = phase_update_bdd(dd_man, event);
	  dda = Cudd_bddAnd(dd_man, ddt, ddu);
	  Cudd_Ref(dda);
	  Cudd_RecursiveDeref(dd_man, ddt);
	  Cudd_RecursiveDeref(dd_man, ddu);
	  DdNode* ddA = Cudd_BddToAdd(dd_man, dda);
	  Cudd_Ref(ddA);
	  Cudd_RecursiveDeref(dd_man, dda);
	  DdNode* ddl = Cudd_addConst(dd_man, (*ri).second);
	  Cudd_Ref(ddl);
	  if (to == params.n) {
	    DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddA, ddP);
	    Cudd_Ref(ddm);
	    Cudd_RecursiveDeref(dd_man, ddA);
	    ddA = ddm;
	    ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddC, ddR);
	    Cudd_Ref(ddm);
	    DdNode* ddr = Cudd_addApply(dd_man, Cudd_addTimes, ddl, ddm);
	    Cudd_Ref(ddr);
	    Cudd_RecursiveDeref(dd_man, ddm);
	    ddm = Cudd_addApply(dd_man, Cudd_addPlus, ddr, ddRe);
	    Cudd_Ref(ddm);
	    Cudd_RecursiveDeref(dd_man, ddr);
	    Cudd_RecursiveDeref(dd_man, ddRe);
	    ddRe = ddr;
	  }
	  Cudd_RecursiveDeref(dd_man, ddC);
	  ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddl, ddA);
	  Cudd_Ref(ddt);
	  Cudd_RecursiveDeref(dd_man, ddA);
	  ddA = ddt;
	  Cudd_RecursiveDeref(dd_man, ddl);
	  if (verbosity > 2) {
	    std::cout << std::endl << "Transition matrix for " << event
		      << " phase transition " << from << "->" << to << ':'
		      << std::endl;
	    Cudd_PrintDebug(dd_man, ddA, 2*nvars, 2);
	  }
	  if (event.controllable()) {
	    action_transitions.insert(std::make_pair(&event, ddA));
	  } else {
	    event_transitions.insert(std::make_pair(&event, ddA));
	  }
	}
      } else {
	throw std::logic_error("no phase data for general distribution");
      }
      if (event.controllable()) {
	action_rewards.insert(std::make_pair(&event, ddRe));
	if (verbosity > 2) {
	  std::cout << std::endl << "Reward vector for " << event << ':'
		    << std::endl;
	  Cudd_PrintDebug(dd_man, ddRe, 2*nvars, 2);
	}
      } else {
	DdNode*& ddr = action_rewards[0];
	DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddRe, ddr);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddRe);
	Cudd_RecursiveDeref(dd_man, ddr);
	ddr = dds;
      }
    }
    if (event.controllable()) {
      DdNode* ddU = action_phase_update_bdd(dd_man, &event);
      DdNode* dda = Cudd_bddAnd(dd_man, ddc, ddU);
      Cudd_Ref(dda);
      Cudd_RecursiveDeref(dd_man, ddU);
      DdNode* ddUa = Cudd_BddToAdd(dd_man, dda);
      Cudd_Ref(ddUa);
      Cudd_RecursiveDeref(dd_man, dda);
      action_phase_updates.insert(std::make_pair(&event, ddUa));
      if (verbosity > 2) {
	std::cout << std::endl << "Phase update for " << event << ':'
		  << std::endl;
	Cudd_PrintDebug(dd_man, ddUa, 2*nvars, 2);
      }
    }
    Cudd_RecursiveDeref(dd_man, ddc);
    Cudd_RecursiveDeref(dd_man, ddP);
    Cudd_RecursiveDeref(dd_man, ddR);
  }
  Cudd_RecursiveDeref(dd_man, state_id);
  if (sv_cube != 0) {
    Cudd_RecursiveDeref(dd_man, sv_cube);
  }

  /*
   * Adjust action rewards.
   */
  for (std::map<const Event*, DdNode*>::iterator ri = action_rewards.begin();
       ri != action_rewards.end(); ri++) {
    if ((*ri).first == 0) {
      DdNode*& ddr = (*ri).second;
      DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, reward_rates, ddr);
      Cudd_Ref(dds);
      Cudd_RecursiveDeref(dd_man, ddr);
      DdNode* ddC = Cudd_BddToAdd(dd_man, ddng);
      Cudd_Ref(ddC);
      ddr = Cudd_addApply(dd_man, Cudd_addTimes, ddC, dds);
      Cudd_Ref(ddr);
      Cudd_RecursiveDeref(dd_man, ddC);
      Cudd_RecursiveDeref(dd_man, dds);
      if (verbosity > 2) {
	std::cout << std::endl << "Reward vector for <idle>:" << std::endl;
	Cudd_PrintDebug(dd_man, ddr, 2*nvars, 2);
      }
    } else {
      DdNode*& ddr = (*ri).second;
      DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus,
				  action_rewards[0], ddr);
      Cudd_Ref(dds);
      Cudd_RecursiveDeref(dd_man, ddr);
      DdNode* ddf = formula_bdd(dd_man, (*ri).first->condition());
      DdNode* ddc = Cudd_bddAnd(dd_man, ddf, ddng);
      Cudd_Ref(ddc);
      Cudd_RecursiveDeref(dd_man, ddf);
      DdNode* ddC = Cudd_BddToAdd(dd_man, ddc);
      Cudd_Ref(ddC);
      Cudd_RecursiveDeref(dd_man, ddc);
      ddr = Cudd_addApply(dd_man, Cudd_addTimes, ddC, dds);
      Cudd_Ref(ddr);
      Cudd_RecursiveDeref(dd_man, ddC);
      Cudd_RecursiveDeref(dd_man, dds);
      if (verbosity > 2) {
	std::cout << std::endl << "Reward vector for " << *(*ri).first << ':'
		  << std::endl;
	Cudd_PrintDebug(dd_man, ddr, 2*nvars, 2);
      }
    }
  }
}


/* ====================================================================== */
/* join_transition_mtbdds */

/*
 * Joins the transition MTBDDs associated with a specific event.
 */
static void
join_transition_mtbdds(DdManager* dd_man,
		       std::multimap<const Event*, DdNode*>& transitions) {
  const Event* last_event = 0;
  DdNode* ddT = Cudd_ReadZero(dd_man);
  Cudd_Ref(ddT);
  for (std::multimap<const Event*, DdNode*>::const_iterator ei =
	 transitions.begin();
       ei != transitions.end(); ei++) {
    const Event* event = (*ei).first;
    if (last_event != 0 && event != last_event) {
      transitions.erase(last_event);
      transitions.insert(std::make_pair(last_event, ddT));
      ddT = Cudd_ReadZero(dd_man);
      Cudd_Ref(ddT);
      ei = transitions.find(event);
    }
    DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddT, (*ei).second);
    Cudd_Ref(dds);
    Cudd_RecursiveDeref(dd_man, ddT);
    Cudd_RecursiveDeref(dd_man, (*ei).second);
    ddT = dds;
    last_event = event;
  }
  if (last_event != 0) {
    transitions.erase(last_event);
    transitions.insert(std::make_pair(last_event, ddT));
  } else {
    Cudd_RecursiveDeref(dd_man, ddT);
  }
}


/* ====================================================================== */
/* Value iteration. */

/*
 * Returns a policy for the current problem generated using value
 * iteration.
 */
static DdNode* value_iteration(std::vector<const Event*>& policy_actions,
			       DdManager* dd_man, const Problem& problem,
			       DdNode* ddng, DdNode* col_cube,
			       double tolerance, double gamma) {
  /*
   * Precompute variable permutations.
   */
  int* row_to_col = new int[2*nvars];
  for (int i = 0; i < nvars; i++) {
    row_to_col[2*i] = 2*i + 1;
    row_to_col[2*i + 1] = 2*i + 1;
  }

  /*
   * Collect actions.
   */
  std::set<const Event*> actions;
  actions.insert(0);
  for (std::map<const Event*, DdNode*>::const_iterator ai =
	 action_transitions.begin();
       ai != action_transitions.end(); ai++) {
    actions.insert((*ai).first);
  }

  /*
   * Construct action conditions, action value filters, and initial policy.
   */
  std::map<const Event*, DdNode*> filters;
  std::map<const Event*, DdNode*> policy;
  for (std::set<const Event*, DdNode*>::const_iterator ai = actions.begin();
       ai != actions.end(); ai++) {
    if (*ai != 0) {
      DdNode* ddc = formula_bdd(dd_man, (*ai)->condition());
      DdNode* ddt = Cudd_bddAnd(dd_man, ddc, ddng);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, ddc);
      DdNode* ddn = Cudd_Not(ddt);
      Cudd_Ref(ddn);
      Cudd_RecursiveDeref(dd_man, ddt);
      ddc = Cudd_BddToAdd(dd_man, ddn);
      Cudd_Ref(ddc);
      Cudd_RecursiveDeref(dd_man, ddn);
      ddn = Cudd_ReadPlusInfinity(dd_man);
      Cudd_Ref(ddn);
      ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddc, ddn);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, ddc);
      Cudd_RecursiveDeref(dd_man, ddn);
      filters.insert(std::make_pair(*ai, ddt));
      DdNode* ddp = Cudd_ReadLogicZero(dd_man);
      Cudd_Ref(ddp);
      policy.insert(std::make_pair(*ai, ddp));
    }
  }

  /*
   * Iterate until value function converges.
   */
  if (verbosity > 0) {
    std::cout << "Value iteration";
  }
  DdNode* ddg = Cudd_addConst(dd_man, gamma);
  Cudd_Ref(ddg);
  DdNode* ddV = Cudd_ReadZero(dd_man);
  Cudd_Ref(ddV);
  if (verbosity > 2) {
    std::cout << std::endl << "V 0:" << std::endl;
    Cudd_PrintDebug(dd_man, ddV, 2*nvars, 2);
  }
  int max_value_size = Cudd_DagSize(ddV);
  int max_prod_size = 0;
  size_t iters = 0;
  bool done = false;
  while (!done) {
    iters++;
    if (verbosity == 1) {
      if (iters % 1000 == 0) {
	std::cout << ':';
      } else if (iters % 100 == 0) {
	std::cout << '.';
      }
    }
    DdNode* ddVp = Cudd_addPermute(dd_man, ddV, row_to_col);
    Cudd_Ref(ddVp);
    DdNode* ddM = 0;
    for (std::set<const Event*>::const_iterator ai = actions.begin();
	 ai != actions.end(); ai++) {
      DdNode* ddE = Cudd_ReadZero(dd_man);
      Cudd_Ref(ddE);
      if (!event_transitions.empty()) {
	DdNode* ddU = action_phase_updates[*ai];
	for (std::multimap<const Event*, DdNode*>::reverse_iterator ei =
	       event_transitions.rbegin();
	     ei != event_transitions.rend(); ei++) {
	  DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes,
				      ddU, (*ei).second);
	  Cudd_Ref(ddm);
	  DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddm, ddVp);
	  Cudd_Ref(ddt);
	  Cudd_RecursiveDeref(dd_man, ddm);
	  max_prod_size = std::max(max_prod_size, Cudd_DagSize(ddt));
	  DdNode* dds = Cudd_addExistAbstract(dd_man, ddt, col_cube);
	  Cudd_Ref(dds);
	  Cudd_RecursiveDeref(dd_man, ddt);
	  ddt = Cudd_addApply(dd_man, Cudd_addPlus, dds, ddE);
	  Cudd_Ref(ddt);
	  Cudd_RecursiveDeref(dd_man, dds);
	  Cudd_RecursiveDeref(dd_man, ddE);
	  ddE = ddt;
	}
      }
      std::pair<trans_iter, trans_iter> bounds =
	action_transitions.equal_range(*ai);
      for (trans_iter ti = bounds.first; ti != bounds.second; ti++) {
	DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, (*ti).second, ddVp);
	Cudd_Ref(ddt);
	max_prod_size = std::max(max_prod_size, Cudd_DagSize(ddt));
	DdNode* dds = Cudd_addExistAbstract(dd_man, ddt, col_cube);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddt);
	ddt = Cudd_addApply(dd_man, Cudd_addPlus, dds, ddE);
	Cudd_Ref(ddt);
	Cudd_RecursiveDeref(dd_man, dds);
	Cudd_RecursiveDeref(dd_man, ddE);
	ddE = ddt;
      }
      DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddg, ddE);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddE);
      DdNode* ddt = Cudd_addApply(dd_man, Cudd_addPlus,
				  action_rewards[*ai], ddm);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, ddm);
      ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddt, holding_times[*ai]);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddt);
      if (*ai != 0) {
	ddE = Cudd_addApply(dd_man, Cudd_addMinus, ddm, filters[*ai]);
	Cudd_Ref(ddE);
	Cudd_RecursiveDeref(dd_man, ddm);
      } else {
	ddE = ddm;
      }
      if (verbosity > 3) {
	std::cout << std::endl << "value of ";
	if (*ai != 0) {
	  std::cout << **ai << ':' << std::endl;
	} else {
	  std::cout << "<idle>:" << std::endl;
	}
	Cudd_PrintDebug(dd_man, ddE, 2*nvars, 2);
      }
      if (*ai != 0) {
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addMinus, ddE, ddM);
	Cudd_Ref(ddm);
	DdNode* ddt = Cudd_addBddStrictThreshold(dd_man, ddm, 0);
	Cudd_Ref(ddt);
	DdNode*& ddp = policy[*ai];
	Cudd_RecursiveDeref(dd_man, ddp);
	ddp  = ddt;
	Cudd_RecursiveDeref(dd_man, ddm);
	DdNode* ddn = Cudd_Not(ddt);
	Cudd_Ref(ddn);
	if (ddn != Cudd_ReadOne(dd_man)) {
	  for (std::map<const Event*, DdNode*>::const_iterator pi =
		 policy.begin();
	       (*pi).first != *ai; pi++) {
	    DdNode*& ddj = policy[(*pi).first];
	    DdNode* dda = Cudd_bddAnd(dd_man, ddn, ddj);
	    Cudd_Ref(dda);
	    Cudd_RecursiveDeref(dd_man, ddj);
	    ddj = dda;
	  }
	}
	Cudd_RecursiveDeref(dd_man, ddn);
	ddm = Cudd_addApply(dd_man, Cudd_addMaximum, ddE, ddM);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddE);
	Cudd_RecursiveDeref(dd_man, ddM);
	ddM = ddm;
      } else {
	if (ddM != 0) {
	  throw std::logic_error("ddM is non-zero");
	}
	ddM = ddE;
      }
      if (verbosity > 3) {
	std::cout << "current max values:" << std::endl;
	Cudd_PrintDebug(dd_man, ddM, 2*nvars, 2);
      }
    }
    Cudd_RecursiveDeref(dd_man, ddVp);
    ddVp = ddM;
    if (verbosity > 2) {
      std::cout << "V " << iters << ':' << std::endl;
      Cudd_PrintDebug(dd_man, ddVp, 2*nvars, 2);
    }
    done = (Cudd_EqualSupNorm(dd_man, ddV, ddVp, tolerance, 0) == 1);
    Cudd_RecursiveDeref(dd_man, ddV);
    ddV = ddVp;
    max_value_size = std::max(max_value_size, Cudd_DagSize(ddV));
  }
  Cudd_RecursiveDeref(dd_man, ddg);
  if (verbosity == 1) {
    std::cout << ' ' << iters << " iterations." << std::endl;
  }
  if (verbosity > 0) {
    std::cout << std::endl
	      << "maximum size of value MTBDD: " << max_value_size << std::endl
	      << "maximum size of product MTBDD: " << max_prod_size
	      << std::endl;
    DdNode* dds = initial_state_mtbdd(dd_man, problem);
    DdNode* ddv = Cudd_addApply(dd_man, Cudd_addTimes, dds, ddV);
    Cudd_Ref(ddv);
    Cudd_RecursiveDeref(dd_man, dds);
    std::cout << std::endl << "initial state value: " << std::endl;
    Cudd_PrintDebug(dd_man, ddv, 2*nvars, 2);
    Cudd_RecursiveDeref(dd_man, ddv);
  }
  Cudd_RecursiveDeref(dd_man, ddV);
  for (std::map<const Event*, DdNode*>::const_iterator ai = filters.begin();
       ai != filters.end(); ai++) {
    Cudd_RecursiveDeref(dd_man, (*ai).second);
  }
  delete row_to_col;

  /*
   * Construct single policy MTBDD.
   */
  DdNode* ddP = Cudd_ReadZero(dd_man);
  Cudd_Ref(ddP);
  for (std::map<const Event*, DdNode*>::const_iterator ai = policy.begin();
       ai != policy.end(); ai++) {
    if (verbosity > 3) {
      std::cout << "condition for " << *(*ai).first << ':' << std::endl;
      Cudd_PrintDebug(dd_man, (*ai).second, 2*nvars, 2);
    }
    if ((*ai).second != Cudd_ReadLogicZero(dd_man)) {
      policy_actions.push_back((*ai).first);
      DdNode* ddp = Cudd_BddToAdd(dd_man, (*ai).second);
      Cudd_Ref(ddp);
      DdNode* ddi = Cudd_addConst(dd_man, policy_actions.size());
      Cudd_Ref(ddi);
      DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddi, ddp);
      Cudd_Ref(ddt);
      Cudd_RecursiveDeref(dd_man, ddi);
      Cudd_RecursiveDeref(dd_man, ddp);
      ddp = Cudd_addApply(dd_man, Cudd_addPlus, ddt, ddP);
      Cudd_Ref(ddp);
      Cudd_RecursiveDeref(dd_man, ddt);
      Cudd_RecursiveDeref(dd_man, ddP);
      ddP = ddp;
    }
    Cudd_RecursiveDeref(dd_man, (*ai).second);
  }
  return ddP;
}


/* ====================================================================== */
/* uniformization */

/*
 * Returns the uniformization constant for the action transition rate
 * matrices and uniformizes the transition rate matrices.
 */
static double uniformization(DdManager* dd_man, DdNode* col_cube,
			     double alpha) {
  /*
   * Compute the row sums for each action transition rate matrix and
   * find the maximum element.
   */
  DdNode** row_variables = new DdNode*[nvars - aux_vars];
  DdNode** col_variables = new DdNode*[nvars - aux_vars];
  for (int i = aux_vars; i < nvars; i++) {
    row_variables[i - aux_vars] = add_var(dd_man, i);
    Cudd_Ref(row_variables[i - aux_vars]);
    col_variables[i - aux_vars] = add_var(dd_man, i, true);
    Cudd_Ref(col_variables[i - aux_vars]);
  }
  DdNode* ddI = Cudd_addXeqy(dd_man, nvars - aux_vars,
			     row_variables, col_variables);
  Cudd_Ref(ddI);
  for (int i = aux_vars; i < nvars; i++) {
    Cudd_RecursiveDeref(dd_man, row_variables[i - aux_vars]);
    Cudd_RecursiveDeref(dd_man, col_variables[i - aux_vars]);
  }
  delete row_variables;
  delete col_variables;
  DdNode* ddN = Cudd_addCmpl(dd_man, ddI);
  Cudd_Ref(ddN);
  DdNode* ddn = Cudd_addConst(dd_man, -1);
  Cudd_Ref(ddn);
  DdNode* ddJ = Cudd_addApply(dd_man, Cudd_addTimes, ddn, ddI);
  Cudd_Ref(ddJ);
  Cudd_RecursiveDeref(dd_man, ddn);
  double q = 0.0;
  for (std::multimap<const Event*, DdNode*>::iterator ei =
	 action_transitions.begin();
       ei != action_transitions.end(); ei++) {
    DdNode* ddQ = Cudd_addApply(dd_man, Cudd_addTimes, ddN, (*ei).second);
    Cudd_Ref(ddQ);
    Cudd_RecursiveDeref(dd_man, (*ei).second);
    DdNode* dds = Cudd_addExistAbstract(dd_man, ddQ, col_cube);
    Cudd_Ref(dds);
    DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddJ, dds);
    Cudd_Ref(ddt);
    (*ei).second = Cudd_addApply(dd_man, Cudd_addPlus, ddt, ddQ);
    Cudd_Ref((*ei).second);
    Cudd_RecursiveDeref(dd_man, ddt);
    Cudd_RecursiveDeref(dd_man, ddQ);
    DdNode* ddq = Cudd_addFindMax(dd_man, dds);
    Cudd_Ref(ddq);
    Cudd_RecursiveDeref(dd_man, dds);
    double qi = Cudd_V(ddq);
    Cudd_RecursiveDeref(dd_man, ddq);
    if (qi > q) {
      q = qi;
    }
  }
  Cudd_RecursiveDeref(dd_man, ddN);
  Cudd_RecursiveDeref(dd_man, ddJ);

  /*
   * Uniformize transition rate matrices.
   */
  DdNode* ddq = Cudd_addConst(dd_man, 1.0/q);
  Cudd_Ref(ddq);
  for (std::multimap<const Event*, DdNode*>::iterator ei =
	 action_transitions.begin();
       ei != action_transitions.end(); ei++) {
    DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddq, (*ei).second);
    Cudd_Ref(ddm);
    Cudd_RecursiveDeref(dd_man, (*ei).second);
    DdNode* ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddI, ddm);
    Cudd_Ref(ddt);
    DdNode* ddp = Cudd_addBddPattern(dd_man, ddt);
    Cudd_Ref(ddp);
    Cudd_RecursiveDeref(dd_man, ddt);
    ddt = Cudd_BddToAdd(dd_man, ddp);
    Cudd_Ref(ddt);
    Cudd_RecursiveDeref(dd_man, ddp);
    (*ei).second = Cudd_addApply(dd_man, Cudd_addPlus, ddt, ddm);
    Cudd_Ref((*ei).second);
    Cudd_RecursiveDeref(dd_man, ddt);
    Cudd_RecursiveDeref(dd_man, ddm);
  }
  Cudd_RecursiveDeref(dd_man, ddq);
  Cudd_RecursiveDeref(dd_man, ddI);

  /*
   * Adjusts the phases for actions, so that transitions for one action
   * always reset the phases of other actions.
   */
  std::map<const Event*, PHData> ph_actions;
  for (std::map<const Event*, PHData>::const_iterator ei = ph_events.begin();
       ei != ph_events.end(); ei++) {
    if ((*ei).first->controllable()) {
      ph_actions.insert(*ei);
    }
  }
  if (!ph_actions.empty()) {
    DdNode** cube_variables = new DdNode*[ph_vars];
    for (std::multimap<const Event*, DdNode*>::iterator ei =
	   action_transitions.begin();
	 ei != action_transitions.end(); ei++) {
      int num_cube_vars = 0;
      for (std::map<const Event*, PHData>::const_iterator ej =
	     ph_actions.begin();
	   ej != ph_actions.end(); ej++) {
	if ((*ei).first != (*ej).first) {
	  const PHData& data = (*ej).second;
	  if (data.low_bit >= 0) {
	    for (int b = data.low_bit; b <= data.high_bit; b++) {
	      DdNode* ddv = add_var(dd_man, b, true);
	      Cudd_Ref(ddv);
	      cube_variables[num_cube_vars++] = ddv;
	    }
	  }
	}
      }
      DdNode* ph_cube = Cudd_addComputeCube(dd_man, cube_variables, 0,
					    num_cube_vars);
      Cudd_Ref(ph_cube);
      DdNode* ddP = Cudd_addExistAbstract(dd_man, (*ei).second, ph_cube);
      Cudd_Ref(ddP);
      for (int i = num_cube_vars - 1; i >= 0; i--) {
	DdNode* ddn = Cudd_addCmpl(dd_man, cube_variables[i]);
	Cudd_Ref(ddn);
	DdNode* dda = Cudd_addApply(dd_man, Cudd_addTimes, ddn, ddP);
	Cudd_Ref(dda);
	Cudd_RecursiveDeref(dd_man, ddn);
	Cudd_RecursiveDeref(dd_man, ddP);
	ddP = dda;
      }
      Cudd_RecursiveDeref(dd_man, (*ei).second);
      (*ei).second = ddP;
      Cudd_RecursiveDeref(dd_man, ph_cube);
      for (int i = 0; i < num_cube_vars; i++) {
	Cudd_RecursiveDeref(dd_man, cube_variables[i]);
      }
    }
    delete cube_variables;
  }

  /*
   * Adjust reward rates.
   */
  // N.B. Does not do the right thing if there are transition rewards
  for (std::map<const Event*, DdNode*>::iterator ri = action_rewards.begin();
       ri != action_rewards.end(); ri++) {
    DdNode* ddq = Cudd_addConst(dd_man, 1.0/(alpha + q));
    Cudd_Ref(ddq);
    DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddq, (*ri).second);
    Cudd_Ref(ddm);
    Cudd_RecursiveDeref(dd_man, ddq);
    Cudd_RecursiveDeref(dd_man, (*ri).second);
    (*ri).second = ddm;
  }

  /*
   * Set holding times to unity.
   */
  for (std::map<const Event*, DdNode*>::iterator hi = holding_times.begin();
       hi != holding_times.end(); hi++) {
    Cudd_RecursiveDeref(dd_man, (*hi).second);
    (*hi).second = Cudd_ReadOne(dd_man);
    Cudd_Ref((*hi).second);
  }

  return q;
}


/* ====================================================================== */
/* mtbdd_solve */

/* Solves the given problem using MTBDDs. */
void mtbdd_solve(DdManager*& dd_man, const Problem& problem,
		 double gamma, double epsilon,
		 PartitionOption partition, FilterOption filter,
		 int moments, int phases, size_t sample_size) {
  itimerval timer = { { 0L, 0L }, { 40000000L, 0L } };
  itimerval stimer;
#ifdef PROFILING
  setitimer(ITIMER_VIRTUAL, &timer, 0);
  getitimer(ITIMER_VIRTUAL, &stimer);
#else
  setitimer(ITIMER_PROF, &timer, 0);
  getitimer(ITIMER_PROF, &stimer);
#endif
  /*
   * Extract the reward function and goal reward.
   */
  Rational goal_reward;
  reward_function = problem.domain().functions().find_function("reward");
  if (reward_function != 0) {
    if (problem.goal_reward() != 0) {
      ValueMap values;
      const Fluent& fluent = problem.goal_reward()->fluent();
      values[&fluent] = 0;
      problem.goal_reward()->affect(values);
      goal_reward = values[&fluent];
    } else {
      goal_reward = 0;
    }
  } else {
    goal_reward = 1;
  }

  /*
   * Collect state variables and assign indices to them.
   */
  nvars = aux_vars = auxiliary_dbn_variables(problem.events());
  nvars += ph_vars = generate_ph_distributions(problem.events(),
					       moments, phases);
  for (EventSet::const_iterator ei = problem.events().begin();
       ei != problem.events().end(); ei++) {
    const Event& event = **ei;
    collect_state_variables(event.effect(), problem.domain());
  }
  for (EffectList::const_iterator ei = problem.init_effects().begin();
       ei != problem.init_effects().end(); ei++) {
    collect_state_variables(**ei, problem.domain());
  }
  ordered_vars.resize(nvars);
  for (int i = 0; i < nvars; i++) {
    var_order.push_back(i);
    ordered_vars[i] = i;
  }
  if (verbosity > 0) {
    std::cout << std::endl << "Number of DD variables: " << nvars
	      << std::endl;
    if (verbosity > 1) {
      for (std::map<int, const Atom*>::const_iterator vi =
	     dynamic_atoms.begin();
	   vi != dynamic_atoms.end(); vi++) {
	std::cout << (*vi).first << '\t' << var_order[(*vi).first] << '\t'
		  << *(*vi).second << std::endl;
      }
      for (std::map<const Event*, PHData>::const_iterator ei =
	     ph_events.begin();
	   ei != ph_events.end(); ei++) {
	const PHData& data = (*ei).second;
	if (data.low_bit >= 0) {
	  for (int b = data.low_bit; b <= data.high_bit; b++) {
	    std::cout << b << '\t' << var_order[b] << "\tphase variable "
		      << b - data.low_bit + 1 << " of " << *(*ei).first
		      << std::endl;
	  }
	}
      }
      for (int i = 0; i < aux_vars; i++) {
	std::cout << i << '\t' << var_order[i] << "\tauxiliary variable"
		  << std::endl;
      }
    }
  }

  /*
   * Initialize CUDD.
   */
  dd_man = Cudd_Init(2*nvars, 0, CUDD_UNIQUE_SLOTS, CUDD_CACHE_SLOTS, 0);

  /*
   * Collect column variables and compute their cube.
   */
  DdNode** col_variables = new DdNode*[nvars - aux_vars];
  for (int i = aux_vars; i < nvars; i++) {
    col_variables[i - aux_vars] = add_var(dd_man, i, true);
    Cudd_Ref(col_variables[i - aux_vars]);
  }
  DdNode* col_cube = Cudd_addComputeCube(dd_man, col_variables, 0,
					 nvars - aux_vars);
  Cudd_Ref(col_cube);
  for (int i = aux_vars; i < nvars; i++) {
    Cudd_RecursiveDeref(dd_man, col_variables[i - aux_vars]);
  }
  delete col_variables;

  /*
   * Construct a BDDs representing goal states.
   */
  DdNode* ddg = formula_bdd(dd_man, problem.goal());
  DdNode* ddng = Cudd_Not(ddg);
  Cudd_Ref(ddng);
  Cudd_RecursiveDeref(dd_man, ddg);
  DdNode* ddgp = formula_bdd(dd_man, problem.goal(), true);

  /*
   * Construct holding time MTBDDs.
   */
  holding_time_mtbdds(dd_man, problem.events(), -log(gamma));

  /*
   * Construct MTBDD for reward rates.
   */
  DdNode* reward_rates =
    reward_rate_mtbdd(dd_man, problem.reward_rates(), ddng);
  if (verbosity > 2) {
    std::cout << std::endl << "MTBDD for reward rates:" << std::endl;
    Cudd_PrintDebug(dd_man, reward_rates, 2*nvars, 2);
  }

  /*
   * Construct MTBDD for goal rewards.
   */
  DdNode* ddgr = goal_reward_mtbdd(dd_man, goal_reward, ddng, ddgp);
  Cudd_RecursiveDeref(dd_man, ddgp);
  if (verbosity > 2) {
    std::cout << std::endl << "MTBDD for goal reward:" << std::endl;
    Cudd_PrintDebug(dd_man, ddgr, 2*nvars, 2);
  }

  /*
   * Construct transition probability and reward MTBDDs for events.
   */
  event_mtbdds(dd_man, problem.events(), ddng, ddgr, reward_rates);
  Cudd_RecursiveDeref(dd_man, ddgr);
  Cudd_RecursiveDeref(dd_man, reward_rates);

  if (partition != FULL_PARTITION) {
    /*
     * Join all transition MTBDDs associated with a specific event.
     */
    join_transition_mtbdds(dd_man, event_transitions);
    join_transition_mtbdds(dd_man, action_transitions);
    if (partition == UNIFORMIZATION || partition == NO_PARTITION) {
      /*
       * Construct a single transition MTBDD for each action.
       */
      DdNode* ddE = Cudd_ReadZero(dd_man);
      Cudd_Ref(ddE);
      for (std::multimap<const Event*, DdNode*>::const_iterator ei =
	     event_transitions.begin();
	   ei != event_transitions.end(); ei++) {
	DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddE, (*ei).second);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddE);
	Cudd_RecursiveDeref(dd_man, (*ei).second);
	ddE = dds;
      }
      event_transitions.clear();
      for (std::multimap<const Event*, DdNode*>::iterator ei =
	     action_transitions.begin();
	   ei != action_transitions.end(); ei++) {
	DdNode* ddU = action_phase_updates[(*ei).first];
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddU, ddE);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddU);
	DdNode* dds = Cudd_addApply(dd_man, Cudd_addPlus, ddm, (*ei).second);
	Cudd_Ref(dds);
	Cudd_RecursiveDeref(dd_man, ddm);
	Cudd_RecursiveDeref(dd_man, (*ei).second);
	(*ei).second = dds;
	if (verbosity > 2) {
	  std::cout << std::endl << "Transition rate matrix for "
		    << *(*ei).first << ':' << std::endl;
	  Cudd_PrintDebug(dd_man, dds, 2*nvars, 2);
	}
      }
      DdNode* ddU = action_phase_updates[0];
      DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddU, ddE);
      Cudd_Ref(ddm);
      Cudd_RecursiveDeref(dd_man, ddU);
      Cudd_RecursiveDeref(dd_man, ddE);
      action_transitions.insert(std::pair<const Event*, DdNode*>(0, ddm));
      if (verbosity > 2) {
	std::cout << std::endl << "Transition rate matrix for idleness:"
		  << std::endl;
	Cudd_PrintDebug(dd_man, ddm, 2*nvars, 2);
      }
      action_phase_updates.clear();
    }
  }

  double discount = 1.0;
  if (partition == UNIFORMIZATION) {
    double q = uniformization(dd_man, col_cube, -log(gamma));
    discount = q/(q - log(gamma));
    if (verbosity > 0) {
      std::cout << std::endl << "uniformization constant is " << q
		<< std::endl;
    }
    if (verbosity > 2) {
      for (std::multimap<const Event*, DdNode*>::const_iterator ei =
	     action_transitions.begin();
	   ei != action_transitions.end(); ei++) {
	std::cout << std::endl << "Transition probability matrix for ";
	if ((*ei).first != 0) {
	  std::cout << *(*ei).first;
	} else {
	  std::cout << "idleness";
	}
	std::cout << ':' << std::endl;
	Cudd_PrintDebug(dd_man, (*ei).second, 2*nvars, 2);
	std::cout << std::endl << "Reward vector for ";
	if ((*ei).first != 0) {
	  std::cout << *(*ei).first;
	} else {
	  std::cout << "idleness";
	}
	std::cout << ':' << std::endl;
	Cudd_PrintDebug(dd_man, action_rewards[(*ei).first], 2*nvars, 2);
      }
    }
  }

  if (filter != NO_FILTER) {
    /*
     * Construct and use phase filter MTBDDs.
     */
    DdNode* ddf = phase_filter_mtbdd(dd_man);
    if (filter == H_FILTER || filter == A_FILTER) {
      for (std::map<const Event*, DdNode*>::iterator hi =
	     holding_times.begin();
	   hi != holding_times.end(); hi++) {
	DdNode*& ddh = (*hi).second;
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddf, ddh);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddh);
	ddh = ddm;
      }
    }
    if (filter == T_FILTER || filter == A_FILTER) {
      DdNode* ddp = phase_filter_mtbdd(dd_man, true);
      for (std::multimap<const Event*, DdNode*>::iterator ei =
	     event_transitions.begin();
	   ei != event_transitions.end(); ei++) {
	DdNode*& ddt = (*ei).second;
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddf, ddt);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddt);
	ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddp, ddm);
	Cudd_Ref(ddt);
	Cudd_RecursiveDeref(dd_man, ddm);
      }
      for (std::multimap<const Event*, DdNode*>::iterator ei =
	     action_transitions.begin();
	   ei != action_transitions.end(); ei++) {
	DdNode*& ddt = (*ei).second;
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddf, ddt);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddt);
	ddt = Cudd_addApply(dd_man, Cudd_addTimes, ddp, ddm);
	Cudd_Ref(ddt);
	Cudd_RecursiveDeref(dd_man, ddm);
      }
      Cudd_RecursiveDeref(dd_man, ddp);
      for (std::map<const Event*, DdNode*>::iterator ri =
	     action_rewards.begin();
	   ri != action_rewards.end(); ri++) {
	DdNode*& ddr = (*ri).second;
	DdNode* ddm = Cudd_addApply(dd_man, Cudd_addTimes, ddf, ddr);
	Cudd_Ref(ddm);
	Cudd_RecursiveDeref(dd_man, ddr);
	ddr = ddm;
      }
    }
    Cudd_RecursiveDeref(dd_man, ddf);
  }

#ifdef PROFILING
  getitimer(ITIMER_VIRTUAL, &timer);
#else
  getitimer(ITIMER_PROF, &timer);
#endif
  long sec = stimer.it_value.tv_sec - timer.it_value.tv_sec;
  long usec = stimer.it_value.tv_usec - timer.it_value.tv_usec;
  double t = std::max(0.0, sec + usec*1e-6);
  std::cout << "Model built in " << t << " seconds." << std::endl;
  stimer = timer;

  /*
   * Generate policy.
   */
  std::vector<const Event*> policy_actions;
  DdNode* ddP = value_iteration(policy_actions, dd_man, problem, ddng,
				col_cube, epsilon*(1.0 - gamma)/(2.0*gamma),
				discount);
  if (verbosity > 1) {
    std::cout << "Policy:" << std::endl;
    Cudd_PrintDebug(dd_man, ddP, 2*nvars, 2);
    for (size_t i = 0; i < policy_actions.size(); i++) {
      std::cout << i + 1 << '\t' << *policy_actions[i] << std::endl;
    }
  }
  Cudd_RecursiveDeref(dd_man, ddng);
  Cudd_RecursiveDeref(dd_man, col_cube);
#ifdef PROFILING
  getitimer(ITIMER_VIRTUAL, &timer);
#else
  getitimer(ITIMER_PROF, &timer);
#endif
  sec = stimer.it_value.tv_sec - timer.it_value.tv_sec;
  usec = stimer.it_value.tv_usec - timer.it_value.tv_usec;
  t = std::max(0.0, sec + usec*1e-6);
  std::cout << "Problem solved in " << t << " seconds." << std::endl;
  if (sample_size > 0) {
    /*
     * Estimate policy value.
     */
    std::map<const Event*, size_t> phase_index;
    for (std::map<const Event*, PHData>::const_iterator pi = ph_events.begin();
	 pi != ph_events.end(); pi++) {
      phase_index.insert(std::make_pair((*pi).first, phase_index.size()));
    }
    Policy policy(dd_man, ddP, policy_actions, ph_parameters, phase_index);
    policy_value(problem, policy, -log(gamma), sample_size);
  } else {
    Cudd_RecursiveDeref(dd_man, ddP);
  }

  /*
   * Clean up.
   */
  state_variables.clear();
  dynamic_atoms.clear();
  ph_parameters.clear();
  ph_events.clear();
  var_order.clear();
  ordered_vars.clear();
  for (std::map<const Event*, DdNode*>::const_iterator ei =
	 holding_times.begin();
       ei != holding_times.end(); ei++) {
    Cudd_RecursiveDeref(dd_man, (*ei).second);
  }
  holding_times.clear();
  for (std::multimap<const Event*, DdNode*>::const_iterator ei =
	 event_transitions.begin();
       ei != event_transitions.end(); ei++) {
    Cudd_RecursiveDeref(dd_man, (*ei).second);
  }
  event_transitions.clear();
  for (std::multimap<const Event*, DdNode*>::const_iterator ei =
	 action_transitions.begin();
       ei != action_transitions.end(); ei++) {
    Cudd_RecursiveDeref(dd_man, (*ei).second);
  }
  action_transitions.clear();
  for (std::map<const Event*, DdNode*>::const_iterator ei =
	 action_phase_updates.begin();
       ei != action_phase_updates.end(); ei++) {
    Cudd_RecursiveDeref(dd_man, (*ei).second);
  }
  action_phase_updates.clear();
  for (std::map<const Event*, DdNode*>::const_iterator ei =
	 action_rewards.begin();
       ei != action_rewards.end(); ei++) {
    Cudd_RecursiveDeref(dd_man, (*ei).second);
  }
  action_rewards.clear();
}


/* ====================================================================== */
/* Policy */

/* Deletes this policy. */
Policy::~Policy() {
  Cudd_RecursiveDeref(dd_man_, mapping_);
}

/* Returns the action selected for the given state. */
const Event* Policy::action(const AtomSet& atoms, const ValueMap& values,
			    const std::vector<size_t>& phases) const {
  DdNode* dds = state_bdd(dd_man_, atoms, phases);
  DdNode* ddS = Cudd_BddToAdd(dd_man_, dds);
  Cudd_Ref(ddS);
  Cudd_RecursiveDeref(dd_man_, dds);
  DdNode* dda = Cudd_addEvalConst(dd_man_, ddS, mapping_);
  Cudd_Ref(dda);
  Cudd_RecursiveDeref(dd_man_, ddS);
  size_t id = int(Cudd_V(dda) + 0.5);
  Cudd_RecursiveDeref(dd_man_, dda);
  return (id > 0) ? actions_[id - 1] : 0;
}
