/* -*-C++-*- */
/*
 * Predicates.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: predicates.h,v 2.2 2005/02/08 11:40:56 lorens Exp $
 */
#ifndef PREDICATES_H
#define PREDICATES_H

#include <config.h>
#include "types.h"
#include <iostream>
#include <map>
#include <set>
#include <string>
#include <vector>


/* ====================================================================== */
/* Predicate */

/*
 * A predicate.
 */
struct Predicate {
  /* Constructs a predicate. */
  explicit Predicate(int index) : index_(index) {}

private:
  /* Predicate index. */
  int index_;

  friend bool operator==(const Predicate& p1, const Predicate& p2);
  friend bool operator<(const Predicate& p1, const Predicate& p2);
  friend std::ostream& operator<<(std::ostream& os, const Predicate& p);
  friend struct PredicateTable;
};

/* Equality operator for predicates. */
inline bool operator==(const Predicate& p1, const Predicate& p2) {
  return p1.index_ == p2.index_;
}

/* Inequality operator for predicates. */
inline bool operator!=(const Predicate& p1, const Predicate& p2) {
  return !(p1 == p2);
}

/* Less-than operator for predicates. */
inline bool operator<(const Predicate& p1, const Predicate& p2) {
  return p1.index_ < p2.index_;
}

/* Output operator for predicates. */
std::ostream& operator<<(std::ostream& os, const Predicate& p);


/* ====================================================================== */
/* PredicateSet */

/* Set of predicate declarations. */
struct PredicateSet : public std::set<Predicate> {
};


/* ====================================================================== */
/* PredicateTable */

/*
 * Predicate table.
 */
struct PredicateTable {
  /* Adds a parameter with the given type to the given predicate. */
  static void add_parameter(const Predicate& predicate, const Type& type);

  /* Returns the name of the given predicate. */
  static const std::string& name(const Predicate& predicate);

  /* Returns the parameter types of the given predicate. */
  static const TypeList& parameters(const Predicate& predicate);

  /* Makes the given predicate dynamic. */
  static void make_dynamic(const Predicate& predicate);

  /* Tests if the given predicate is static. */
  static bool static_predicate(const Predicate& predicate);

  /* Adds a predicate with the given name to this table and returns
     the predicate. */
  const Predicate& add_predicate(const std::string& name);

  /* Returns a pointer to the predicate with the given name, or 0 if
     no predicate with the given name exists. */
  const Predicate* find_predicate(const std::string& name) const;

private:
  /* Predicate names. */
  static std::vector<std::string> names_;
  /* Predicate parameters. */
  static std::vector<TypeList> parameters_;
  /* Static predicates. */
  static PredicateSet static_predicates_;

  /* Mapping of predicate names to predicates. */
  std::map<std::string, Predicate> predicates_;

  friend std::ostream& operator<<(std::ostream& os, const PredicateTable& t);
  friend std::ostream& operator<<(std::ostream& os, const Predicate& p);
};

/* Output operator for predicate tables. */
std::ostream& operator<<(std::ostream& os, const PredicateTable& t);


#endif /* PREDICATES_H */
