/*
 * States.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: states.cc,v 2.2 2005/02/08 11:44:27 lorens Exp $
 */
#include "states.h"
#include "effects.h"
#include "rng.h"
#include <stdexcept>


/* Verbosity level. */
extern int verbosity;


/* ====================================================================== */
/* State */

/*
 * An extended state.
 */
struct ExtendedState : public State {
  /* Constructs an extended state. */
  ExtendedState(const AtomSet& atoms, const ValueMap& values,
		const std::vector<size_t>& phases, double dt);

  /* Returns the time spent in the previous state. */
  virtual double dt() const { return dt_; }

  /* Returns a successor of this state. */
  virtual const State* next(const Problem& problem,
			    const EventList& exogenous_events,
			    const Policy& policy) const;

protected:
  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

private:
  /* A mapping from events to lifetimes. */
  struct LifetimeMap : public std::map<const Event*, double> {
  };

  /* Values of phase variables. */
  std::vector<size_t> phases_;
  /* Time spent in the previous state. */
  double dt_;
  /* Sampled lifetimes for enabled events. */
  LifetimeMap lifetimes_;
  /* Event with shortest sampled lifetime. */
  const Event* trigger_;
  /* Next phase if PHEvent has the shortest sampled lifetime. */
  size_t next_phase_;
  /* The shortest sampled lifetime. */
  double trigger_time_;

  friend const State* State::next(const Problem& problem,
				  const EventList& exogenous_events,
				  const Policy& policy) const;
};


/* Constructs an initial state for the given problem. */
State::State(const Problem& problem) {
  for (AtomSet::const_iterator ai = problem.init_atoms().begin();
       ai != problem.init_atoms().end(); ai++) {
    if (!PredicateTable::static_predicate((*ai)->predicate())) {
      atoms_.insert(*ai);
    }
  }
  for (ValueMap::const_iterator vi = problem.init_values().begin();
       vi != problem.init_values().end(); vi++) {
    Function function = (*vi).first->function();
    if (!FunctionTable::static_function(function)) {
      values_.insert(*vi);
    }
  }
  for (EffectList::const_iterator ei = problem.init_effects().begin();
       ei != problem.init_effects().end(); ei++) {
    AtomList adds;
    AtomList deletes;
    UpdateList updates;
    (*ei)->state_change(adds, deletes, updates, atoms_, values_);
    atoms_.insert(adds.begin(), adds.end());
    for (UpdateList::const_iterator ui = updates.begin();
	 ui != updates.end(); ui++) {
      (*ui)->affect(values_);
    }
  }
  goal_ = problem.goal().holds(atoms_, values_);
  if (goal()) {
    const Fluent& goal_achieved_fluent =
      Fluent::make(problem.domain().goal_achieved(), TermList());
    values_[&goal_achieved_fluent] = 1;
    if (problem.goal_reward() != 0) {
      problem.goal_reward()->affect(values_);
      reward_ = values_[&problem.goal_reward()->fluent()];
    }
  }
}


/* Returns a successor of this state. */
const State* State::next(const Problem& problem,
			 const EventList& exogenous_events,
			 const Policy& policy) const {
  EventList events;
  for (EventList::const_iterator ei = exogenous_events.begin();
       ei != exogenous_events.end(); ei++) {
    if ((*ei)->condition().holds(atoms(), values())) {
      events.push_back(*ei);
    }
  }
  std::vector<size_t> phases(policy.phase_index().size(), 0);
  const Event* action = policy.action(atoms(), values(), phases);
  if (action != 0) {
    if (!action->condition().holds(atoms(), values())) {
      throw std::logic_error("disabled action selected by policy");
    }
    events.push_back(action);
  }
  ExtendedState* next_state = new ExtendedState(atoms(), values(),
						phases, 0.0);
  int streak = 1;
  double tie_breaker = -1.0;
  for (EventList::const_iterator ei = events.begin();
       ei != events.end(); ei++) {
    const Event* e = *ei;
    double t = e->delay().sample();
    next_state->lifetimes_.insert(std::make_pair(e, t));
    if (t == next_state->trigger_time_) {
      streak++;
      if (tie_breaker < 0.0) {
	tie_breaker = genrand_real2();
      }
    } else if (t < next_state->trigger_time_) {
      streak = 1;
    }
    if (t < next_state->trigger_time_
	|| (t == next_state->trigger_time_ && tie_breaker < 1.0/streak)) {
      next_state->trigger_ = e;
      next_state->next_phase_ = 0;
      next_state->trigger_time_ = t;
    }
    PHEventMap::const_iterator pi = policy.ph_parameters().find(e->name());
    if (pi != policy.ph_parameters().end()) {
      const PHParameters& params = (*pi).second;
      for (TransitionRateMap::const_iterator ri = params.rates.begin();
	   ri != params.rates.end(); ri++) {
	if ((*ri).first.first == 0 && (*ri).first.second != params.n) {
	  double t = Distribution::EXP1.sample()/(*ri).second;
	  if (t == next_state->trigger_time_) {
	    streak++;
	    if (tie_breaker < 0.0) {
	      tie_breaker = genrand_real2();
	    }
	  } else if (t < next_state->trigger_time_) {
	    streak = 1;
	  }
	  if (t < next_state->trigger_time_
	      || (t == next_state->trigger_time_
		  && tie_breaker < 1.0/streak)) {
	    next_state->trigger_ = e;
	    next_state->next_phase_ = (*ri).first.second;
	    next_state->trigger_time_ = t;
	  }
	}
      }
    }
  }
  const State* result = next_state->next(problem, exogenous_events, policy);
  delete next_state;
  return result;
}


/* Prints this object on the given stream. */
void State::print(std::ostream& os) const {
  AtomSet::const_iterator ai = atoms().begin();
  if (ai != atoms().end()) {
    os << **ai;
    for (ai++; ai != atoms().end(); ai++) {
      os << " & " << **ai;
    }
  }
}


/* Output operator for states. */
std::ostream& operator<<(std::ostream& os, const State& s) {
  s.print(os);
  return os;
}


/* Constructs an extended state. */
ExtendedState::ExtendedState(const AtomSet& atoms, const ValueMap& values,
			     const std::vector<size_t>& phases, double dt)
  : State(atoms, values, 0), phases_(phases), dt_(dt), trigger_(0),
    next_phase_(0), trigger_time_(HUGE_VAL) {}


/* Returns a successor of this state. */
const State* ExtendedState::next(const Problem& problem,
				 const EventList& exogenous_events,
				 const Policy& policy) const {
  if (verbosity > 2) {
    for (LifetimeMap::const_iterator ti = lifetimes_.begin();
	 ti != lifetimes_.end(); ti++) {
      std::cout << "  " << *(*ti).first << " @ " << (*ti).second << std::endl;
    }
    if (next_phase_ > 0) {
      std::cout << "  " << *trigger_ << " phase change to " << next_phase_
		<< " @ " << trigger_time_ << std::endl;
    }
  }
  ExtendedState* next_state = new ExtendedState(atoms(), values(), phases_,
						trigger_time_);
  const Event* old_action = policy.action(atoms(), values(), phases_);
  for (std::map<const Event*, size_t>::const_iterator pi =
	 policy.phase_index().begin();
       pi != policy.phase_index().end(); pi++) {
    if ((*pi).first->controllable() && (*pi).first != old_action) {
      next_state->phases_[(*pi).second] = 0;
    }
  }
  if (next_phase_ > 0) {
    std::map<const Event*, size_t>::const_iterator phi =
      policy.phase_index().find(trigger_);
    if (phi == policy.phase_index().end()) {
      throw std::logic_error("phase has no index");
    }
    int ph_phase = (*phi).second;
    /* Update phase according to trigger. */
    next_state->phases_[ph_phase] = next_phase_;
    /* Find action for next state. */
    const Event* action = policy.action(atoms(), values(),
					next_state->phases_);
    int streak = 1;
    double tie_breaker = -1.0;
    if (action != 0) {
      if (!action->condition().holds(atoms(), values())) {
	throw std::logic_error("disabled action selected by policy");
      }
      if (action != old_action) {
	double t =  action->delay().sample();
	next_state->lifetimes_.insert(std::make_pair(action, t));
	next_state->trigger_ = action;
	next_state->next_phase_ = 0;
	next_state->trigger_time_ = t;
	PHEventMap::const_iterator pi =
	  policy.ph_parameters().find(action->name());
	if (pi != policy.ph_parameters().end()) {
	  const PHParameters& params = (*pi).second;
	  for (TransitionRateMap::const_iterator ri = params.rates.begin();
	       ri != params.rates.end(); ri++) {
	    if ((*ri).first.first == 0 && (*ri).first.second != params.n) {
	      double t = Distribution::EXP1.sample()/(*ri).second;
	      if (t == next_state->trigger_time_) {
		streak++;
		if (tie_breaker < 0.0) {
		  tie_breaker = genrand_real2();
		}
	      } else if (t < next_state->trigger_time_) {
		streak = 1;
	      }
	      if (t < next_state->trigger_time_
		  || (t == next_state->trigger_time_
		      && tie_breaker < 1.0/streak)) {
		next_state->trigger_ = action;
		next_state->next_phase_ = (*ri).first.second;
		next_state->trigger_time_ = t;
	      }
	    }
	  }
	}
      }
    }
    for (LifetimeMap::const_iterator ei = lifetimes_.begin();
	 ei != lifetimes_.end(); ei++) {
      const Event* e = (*ei).first;
      if (e->controllable() && e != action) {
	continue;
      }
      double t = (*ei).second - trigger_time_;
      next_state->lifetimes_.insert(std::make_pair(e, t));
      if (t == next_state->trigger_time_) {
	streak++;
	if (tie_breaker < 0.0) {
	  tie_breaker = genrand_real2();
	}
      } else if (t < next_state->trigger_time_) {
	streak = 1;
      }
      if (t < next_state->trigger_time_
	  || (t == next_state->trigger_time_ && tie_breaker < 1.0/streak)) {
	next_state->trigger_ = e;
	next_state->next_phase_ = 0;
	next_state->trigger_time_ = t;
      }
      PHEventMap::const_iterator pi = policy.ph_parameters().find(e->name());
      if (pi != policy.ph_parameters().end()) {
	const PHParameters& params = (*pi).second;
	std::map<const Event*, size_t>::const_iterator phi =
	  policy.phase_index().find(e);
	if (phi == policy.phase_index().end()) {
	  throw std::logic_error("phase has no index");
	}
	size_t ph = next_state->phases_[(*phi).second];
	for (TransitionRateMap::const_iterator ri = params.rates.begin();
	     ri != params.rates.end(); ri++) {
	  if ((*ri).first.first == ph && (*ri).first.second != params.n) {
	    double t = Distribution::EXP1.sample()/(*ri).second;
	    if (t == next_state->trigger_time_) {
	      streak++;
	      if (tie_breaker < 0.0) {
		tie_breaker = genrand_real2();
	      }
	    } else if (t < next_state->trigger_time_) {
	      streak = 1;
	    }
	    if (t < next_state->trigger_time_
		|| (t == next_state->trigger_time_
		    && tie_breaker < 1.0/streak)) {
	      next_state->trigger_ = e;
	      next_state->next_phase_ = (*ri).first.second;
	      next_state->trigger_time_ = t;
	    }
	  }
	}
      }
    }
  } else { /* next_phase == 0 */
    Function reward_function =
      *problem.domain().functions().find_function("reward");
    const Fluent& reward_fluent = Fluent::make(reward_function, TermList());
    next_state->reward_ = -next_state->values_[&reward_fluent];
    trigger_->affect(next_state->atoms_, next_state->values_);
    next_state->goal_ = problem.goal().holds(next_state->atoms(),
					     next_state->values());
    if (next_state->goal()) {
      const Fluent& goal_achieved_fluent =
	Fluent::make(problem.domain().goal_achieved(), TermList());
      next_state->values_[&goal_achieved_fluent] = 1;
      if (problem.goal_reward() != 0) {
	problem.goal_reward()->affect(next_state->values_);
      }
    }
    next_state->reward_ =
      next_state->reward_ + next_state->values_[&reward_fluent];
    std::map<const Event*, size_t>::const_iterator pi =
      policy.phase_index().find(trigger_);
    if (pi != policy.phase_index().end()) {
      next_state->phases_[(*pi).second] = 0;
    }
    EventList events;
    for (EventList::const_iterator ei = exogenous_events.begin();
	 ei != exogenous_events.end(); ei++) {
      if ((*ei)->condition().holds(next_state->atoms(),
				   next_state->values())) {
	events.push_back(*ei);
      } else {
	std::map<const Event*, size_t>::const_iterator pi =
	  policy.phase_index().find(*ei);
	if (pi != policy.phase_index().end()) {
	  next_state->phases_[(*pi).second] = 0;
	}
      }
    }
    const Event* action = policy.action(next_state->atoms(),
					next_state->values(),
					next_state->phases_);
    if (action != 0) {
      if (!action->condition().holds(next_state->atoms(),
				     next_state->values())) {
	throw std::logic_error("disabled action selected by policy");
      }
      events.push_back(action);
    }
    int streak = 1;
    double tie_breaker = -1.0;
    for (EventList::const_iterator ei = events.begin();
	 ei != events.end(); ei++) {
      const Event* e = *ei;
      double t;
      if (e == trigger_) {
	t = e->delay().sample();
      } else {
	LifetimeMap::const_iterator ti = lifetimes_.find(e);
	if (ti != lifetimes_.end()) {
	  t = (*ti).second - trigger_time_;
	} else {
	  t = e->delay().sample();
	}
      }
      next_state->lifetimes_.insert(std::make_pair(e, t));
      if (t == next_state->trigger_time_) {
	streak++;
	if (tie_breaker < 0.0) {
	  tie_breaker = genrand_real2();
	}
      } else if (t < next_state->trigger_time_) {
	streak = 1;
      }
      if (t < next_state->trigger_time_
	  || (t == next_state->trigger_time_ && tie_breaker < 1.0/streak)) {
	next_state->trigger_ = e;
	next_state->next_phase_ = 0;
	next_state->trigger_time_ = t;
      }
      PHEventMap::const_iterator pi = policy.ph_parameters().find(e->name());
      if (pi != policy.ph_parameters().end()) {
	const PHParameters& params = (*pi).second;
	std::map<const Event*, size_t>::const_iterator phi =
	  policy.phase_index().find(e);
	if (phi == policy.phase_index().end()) {
	  throw std::logic_error("phase has no index");
	}
	size_t ph = next_state->phases_[(*phi).second];
	for (TransitionRateMap::const_iterator ri = params.rates.begin();
	     ri != params.rates.end(); ri++) {
	  if ((*ri).first.first == ph && (*ri).first.second != params.n) {
	    double t = Distribution::EXP1.sample()/(*ri).second;
	    if (t == next_state->trigger_time_) {
	      streak++;
	      if (tie_breaker < 0.0) {
		tie_breaker = genrand_real2();
	      }
	    } else if (t < next_state->trigger_time_) {
	      streak = 1;
	    }
	    if (t < next_state->trigger_time_
		|| (t == next_state->trigger_time_
		    && tie_breaker < 1.0/streak)) {
	      next_state->trigger_ = e;
	      next_state->next_phase_ = (*ri).first.second;
	      next_state->trigger_time_ = t;
	    }
	  }
	}
      }
    }
  }
  return next_state;
}


/* Prints this object on the given stream. */
void ExtendedState::print(std::ostream& os) const {
  AtomSet::const_iterator ai = atoms().begin();
  if (ai != atoms().end()) {
    os << **ai;
    for (ai++; ai != atoms().end(); ai++) {
      os << " & " << **ai;
    }
  }
  for (std::vector<size_t>::const_iterator pi = phases_.begin();
       pi != phases_.end(); pi++) {
    os << ' ' << *pi;
  }
}


/* ====================================================================== */
/* policy_value */

/*
 * Generates a single sample execution path and returns the reward.
 */
static double single_execution(const Problem& problem,
			       const EventList& exogenous_events,
			       const Policy& policy, double alpha) {
  double time = 0.0;
  const State* state = new State(problem);
  double reward = state->transition_reward().double_value();
  if (verbosity > 1) {
    std::cout << "Initial state @ " << time << ':' << std::endl << *state
	      << std::endl;
  }
  double max_reward_rate = 0.0;
  for (RewardRateList::const_iterator ri = problem.reward_rates().begin();
       ri != problem.reward_rates().end(); ri++) {
    max_reward_rate += std::abs((*ri).second.double_value());
  }
  double discount = 1.0;
  while (!state->goal() && max_reward_rate*discount > 1e-6) {
    double reward_rate = 0.0;
    for (RewardRateList::const_iterator ri = problem.reward_rates().begin();
	 ri != problem.reward_rates().end(); ri++) {
      if ((*ri).first->holds(state->atoms(), state->values())) {
	reward_rate += (*ri).second.double_value();
      }
    }
    reward_rate /= alpha;
    const State* next_state = state->next(problem, exogenous_events, policy);
    delete state;
    state = next_state;
    time += state->dt();
    double new_discount = exp(-alpha*time);
    reward += ((state->transition_reward().double_value()
		- reward_rate)*new_discount
	       + reward_rate*discount);
    discount = new_discount;
    if (verbosity > 1) {
      std::cout << "State @ " << time << ':' << std::endl << *state
		<< std::endl;
    }
  }
  delete state;
  return reward;
}


/*
 * Estimates the value of the given policy in the initial state of the
 * given problem using sampling.
 */
double policy_value(const Problem& problem, const Policy& policy,
		    double alpha, size_t n) {
  if (verbosity == 1) {
    std::cout << "Estimating policy value";
  }
  EventList exogenous_events;
  for (EventSet::const_iterator ei = problem.events().begin();
       ei != problem.events().end(); ei++) {
    if (!(*ei)->controllable()) {
      exogenous_events.push_back(*ei);
    }
  }
  std::vector<double> rewards;
  double mean = 0.0;
  for (size_t i = 0; i < n; i++) {
    double r = single_execution(problem, exogenous_events, policy, alpha);
    rewards.push_back(r);
    mean += r;
    if (verbosity > 0) {
      if (verbosity == 1) {
	if ((i + 1) % 1000 == 0) {
	  std::cout << ':';
	} else if ((i + 1) % 100 == 0) {
	  std::cout << '.';
	}
      } else {
	std::cout << "sample " << i + 1 << ": " << r
		  << "\t(" << mean/(i + 1) << ")" << std::endl;
      }
    }
  }
  if (verbosity == 1) {
    std::cout << std::endl;
  }
  mean /= n;
  double var = 0.0;
  if (n > 1) {
    for (std::vector<double>::const_iterator ri = rewards.begin();
	 ri != rewards.end(); ri++) {
      double x = *ri - mean;
      var += x*x;
    }
    var /= n - 1;
  }
  std::cout << "mean: " << mean << std::endl;
  std::cout << "dev:  " << sqrt(var) << std::endl;
  return mean;
}
