/* -*-C++-*- */
/*
 * States.
 *
 * Copyright (C) 2004, 2005 Carnegie Mellon University
 *
 * This file is part of Tempastic-DTP.
 *
 * Tempastic-DTP is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * Tempastic-DTP is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempastic-DTP; if not, write to the Free Software
 * Foundation, Inc., #59 Temple Place, Suite 330, Boston, MA
 * 02111-1307 USA
 *
 * $Id: states.h,v 2.2 2005/02/08 11:45:11 lorens Exp $
 */
#ifndef STATES_H
#define STATES_H

#include <config.h>
#include "mtbdd.h"
#include "problems.h"
#include "events.h"
#include "formulas.h"
#include "expressions.h"
#include <iostream>


/* ====================================================================== */
/* State */

/*
 * A state.
 */
struct State {
  /* Constructs an initial state for the given problem. */
  explicit State(const Problem& problem);

  /* Deletes this state. */
  virtual ~State() {}

  /* Returns the set of atoms that hold in this state. */
  const AtomSet& atoms() const { return atoms_; }

  /* Returns the fluent values for this state. */
  const ValueMap& values() const { return values_; }

  /* Tests if this is a goal state. */
  bool goal() const { return goal_; }

  /* Returns the reward associated with the last transition. */
  Rational transition_reward() const { return reward_; }

  /* Returns the time spent in the previous state. */
  virtual double dt() const { return 0.0; }

  /* Returns a successor of this state. */
  virtual const State* next(const Problem& problem,
			    const EventList& exogenous_events,
			    const Policy& policy) const;

protected:
  /* Atoms that hold in this state. */
  AtomSet atoms_;
  /* Fluent values in this state. */
  ValueMap values_;
  /* Whether this is a goal state. */
  bool goal_;
  /* Reward associated with the last transition. */
  Rational reward_;

  /* Constructs a state. */
  State(const AtomSet& atoms, const ValueMap& values, const Rational& reward)
    : atoms_(atoms), values_(values), goal_(false), reward_(reward) {}

  /* Prints this object on the given stream. */
  virtual void print(std::ostream& os) const;

  friend std::ostream& operator<<(std::ostream& os, const State& s);
};

/* Output operator for states. */
std::ostream& operator<<(std::ostream& os, const State& s);


/* ====================================================================== */
/* policy_value */

/*
 * Estimates the value of the given policy in the initial state of the
 * given problem using sampling.
 */
double policy_value(const Problem& problem, const Policy& policy,
		    double alpha, size_t n);


#endif /* STATES_H */
